import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, List, Optional, Union, Dict, TypeVar, Type, Tuple
from netaddr import IPAddress, IPNetwork

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration import ServiceParameter
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.message import Message
from cyst.api.environment.physical import PhysicalAccess, PhysicalLocation, PhysicalConnection
from cyst.api.host.service import Service, PassiveService, ActiveService
from cyst.api.logic.action import ActionParameterDomain
from cyst.api.logic.access import Authorization, AccessLevel, AuthenticationTokenType, AuthenticationTokenSecurity,\
    AuthenticationToken, AuthenticationProviderType, AuthenticationProvider, AccessScheme, AuthenticationTarget
from cyst.api.logic.data import Data
from cyst.api.logic.exploit import VulnerableService, ExploitCategory, ExploitLocality, ExploitParameter, ExploitParameterType, Exploit
from cyst.api.network.elements import Connection, Interface, Route
from cyst.api.network.firewall import FirewallRule, FirewallPolicy
from cyst.api.network.session import Session
from cyst.api.network.node import Node
from cyst.api.utils.duration import Duration


ActiveServiceInterfaceType = TypeVar('ActiveServiceInterfaceType')
ConfigurationObjectType = TypeVar('ConfigurationObjectType')
ObjectType = TypeVar('ObjectType')


class EnvironmentConfiguration(ABC):
    """
    This interface is a collection of configuration interfaces for the environment, that are split according to
    their general functionality.
    """

    @property
    @abstractmethod
    def general(self) -> 'GeneralConfiguration':
        """
        General configuration enables retrieval of objects and their configurations. As such it enables manipulation with
        all objects that are present in the simulation run.

        :rtype: GeneralConfiguration
        """

    @property
    @abstractmethod
    def node(self) -> 'NodeConfiguration':
        """
        Node configuration enables creation and manipulation with nodes and routers.

        :rtype: NodeConfiguration
        """

    @property
    @abstractmethod
    def service(self) -> 'ServiceConfiguration':
        """
        Service configuration enables management of passive and active services.

        :rtype: ServiceConfiguration
        """

    @property
    @abstractmethod
    def network(self) -> 'NetworkConfiguration':
        """
        Network configuration enables placing nodes inside the topology and manipulation with connection and sessions.
        Currently, it is mostly additive, but it will include all manipulation options in the future.

        :rtype: NetworkConfiguration
        """

    @property
    @abstractmethod
    def exploit(self) -> 'ExploitConfiguration':
        """
        This interface provides means to manipulate with exploits.

        :rtype: ExploitConfiguration
        """

    @property
    @abstractmethod
    def action(self) -> 'ActionConfiguration':
        """
        Action configuration enables creation of action parameter domains. While the actions are fully declarative in their
        description and their semantics stem from the interpretation given by behavioral models, action parameters enable
        fine-tuning of actions, that is accessible to automatic tools in a uniform manner.

        :rtype: ActionConfiguration
        """

    @property
    @abstractmethod
    def access(self) -> 'AccessConfiguration':
        """
        Access configuration enables manipulation of authentication and authorization primitives, creation of access schemes
        and evaluation of authentication tokens.

        :rtype: AccessConfiguration
        """

    @property
    @abstractmethod
    def physical(self) -> 'PhysicalConfiguration':
        """
        Physical configuration enables manipulation with physical locations, their assets, and users within them.

        :rtype: PhysicalConfiguration
        """

class GeneralConfiguration(ABC):
    """
    General configuration enables retrieval of objects and their configurations. As such it enables manipulation with
    all objects that are present in the simulation run.
    """
    @abstractmethod
    def get_configuration(self) -> List[ConfigItem]:
        """
        Get the entire configuration of the environment in form of a list containing top level configuration objects,
        i.e., nodes, connections, and exploits. There are no inter-object references, they are all resolved, so there
        is a potential duplication.

        :return: A complete environment configuration.
        """

    @abstractmethod
    def save_configuration(self, indent: Optional[int]) -> str:
        """
        Serializes a configuration into a string representation, that can be saved and later loaded and passed to the
        configure() method.

        :param indent: The indentation level for pretty-printing. If not provided, the configuration is serialized onto
                       one line. If indent is set to 0, only line breaks are inserted. If indent is set to 1 or more
                       the text is indented accordingly.
        :type indent: Optional[int]

        :return: A string representation of the environment's configuration.
        """

    @abstractmethod
    def load_configuration(self, config: str) -> List[ConfigItem]:
        """
        Deserializes a string configuration into corresponding configuration items. These are currently not guaranteed
        to work across versions.

        :param config: The serialized configuration.
        :type config: str

        :return: A list of configuration items that can be fed to the configure() function.
        """

    @abstractmethod
    def get_configuration_by_id(self, id: str, configuration_type: Type[ConfigurationObjectType]) -> ConfigurationObjectType:
        """
        Get a ...Config object by ID.

        :param id: The ID of the configuration object.
        :type id: str

        :param configuration_type: A type of the configuration item to get. While not technically necessary, given the
            state of Python's type system, it is there to satisfy static typing inspection.
        :type configuration_type: Type[TypeVar('ConfigurationObjectType')]

        :return: A configuration object with given ID.
        """

    @abstractmethod
    def get_object_by_id(self, id: str, object_type: Type[ObjectType]) -> ObjectType:
        """
        Get this back...
        :param id:
        :param object_type:
        :return:
        """


class NodeConfiguration(ABC):
    """
    Node configuration enables creation and manipulation with nodes and routers.
    """
    @abstractmethod
    def create_node(self, id: str, ip: Union[str, IPAddress] = "", mask: str = "", shell: Service = None) -> Node:
        """
        Create a node within a simulation. The node itself is empty and must be filled with services before it can
        interact with the environment. The same goes for its network connection, unless provided in the constructor.

        :param id: A unique identification of the node within the simulation. Setting a duplicate ID would result in
            an exception.
        :type id: str

        :param ip: An IP address this node should use. Setting this parameter automatically creates the appropriate
            network interface. The address can be passed as a string or as an IPAddress, into which it is internally
            converted from the string.
        :type ip: Union[str, IPAddress]

        :param mask: A network mask to use. This will be used for setting the gateway and routing policy on the node.
            If the mask is not selected, it gets negotiated during a connection with the router (TODO: check).
        :type mask: str

        :param shell: A service that will be acting as a system shell.
        :type shell: Service

        :return: An instance of a node in the simulation.
        """

    @abstractmethod
    def create_router(self, id: str, messaging: EnvironmentMessaging) -> Node:
        """
        Create a router within simulation. The router is a special type of node, which enables message forwarding and
        network partitioning.

        :param id: A unique identification of the node within simulation.  Setting a duplicate ID would result in
            an exception.
        :type id: str

        :param messaging: A reference to EnvironmentMessaging
        :type messaging: EnvironmentMessaging

        :return: An instance of a router in the simulation.
        """

    @abstractmethod
    def create_port(self, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0, id: str = "") -> Interface:
        """
        Create a network port, which can then be used to connect routers to routers.

        :param ip: An IP address of the port. If not set, it will be a port without IP and the routing will have to be
            explicitly specified.
        :type ip: Union[str, IPAddress]

        :param mask: A network mask of the port. It is used to signal the network range serviced by the router.
        :type mask: str

        :param index: An index the port should have. If a value -1 is selected, the interface is assigned a lowest
            possible index.
        :type index: int

        :param id: A unique identification of the interface within the simulation. If no value is provided, the system
                   will generate a unique one.
        :type id: str

        :return: An instance of a network port.
        """
        pass

    @abstractmethod
    def create_interface(self, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0, id: str = "") -> Interface:
        """
        Create a network interface, which can then be used to connect nodes to routers or routers to routers.

        :param ip: An IP address of the interface. If not set, it will act as a DHCP interface for nodes and an IP-less
            port for routers.
        :type ip: Union[str, IPAddress]

        :param mask: A network mask of the interface. In case of a node interface it will be used for setting the
            gateway and routing policy on the node. If the mask is not selected, it gets negotiated during a connection
            with the router. In case of router port, the mask is used to determine the pool for DHCP addresses.
        :type mask: str

        :param index: An index the port should have, once assigned to a node. If a value -1 is selected, the interface
            is assigned a lowest possible index.
        :type index: int

        :param id: A unique identification of the interface within the simulation. If no value is provided, the system
                   will generate a unique one.
        :type id: str

        :return: An instance of a network interface.
        """

    @abstractmethod
    def create_route(self, net: IPNetwork, port: int, metric: int, id: str = "") -> Route:
        """
        Create a route, which determines through which port messages directed to particular network get sent.

        :param net: A destination network, which is being routed through the specified port.
        :type net: IPNetwork

        :param port: A port index. If a nonexistent port is selected, it will raise an exception, when a message will
            be routed through it.
        :type port: int

        :param metric: A metric for the route. Routes with lower metric get selected when there is an overlap in
            their networks. In case of metric equivalence, more specific routes are selected.
        :type metric: int

        :param id: A unique identification of the route within the simulation. If no value is provided, the system
                   will generate a unique one.
        :type id: str

        :return: A route to be used in router.
        """

    @abstractmethod
    def add_interface(self, node: Node, interface: Interface, index: int = -1) -> int:
        """
        Assigns a network interface instance to a node.

        :param node: A node instance to assign the interface to.
        :type node: Node

        :param interface: An interface instance to be assigned.
        :type interface: Interface

        :param index: An index to give the interface. If left at -1, a lowest possible index is given to the interface.
            If an already occupied index is given, an exception is thrown.

        :return: An index, where the interface was assigned.
        """

    @abstractmethod
    def set_interface(self, interface: Interface, ip: Union[str, IPAddress] = "", mask: str = "") -> None:
        """
        Sets parameters of an interface instance.

        :param interface: The instance of the interface.
        :type interface: Interface

        :param ip: A new IP address to use. Setting this to an empty string will effectively erase the IP address from
            the interface. In that case, it can only work in routers with explicit routing.
        :type ip: Union[str, IPAddress]

        :param mask: A new network mask.
        :type mask: str

        :return: None
        """

    @abstractmethod
    def add_service(self, node: Node, *service: Service) -> None:
        """
        Add a service to the node. The engine expects that there aren't two services of the same name on one node. In
        such case, the effects are undefined.

        :param node: An instance of a node to assign the service to.
        :type node: Node

        :param service: One or more service instances to assign.
        :type service: Service

        :return: None
        """

    @abstractmethod
    def remove_service(self, node: Node, *service: Service) -> None:
        """
        Remove a service from the node.

        :param node: An instance of a node to remove the service from.
        :type node: Node

        :param sertice: One or more service instances to remove.
        :type service: Service

        :return: None
        """

    @abstractmethod
    def set_shell(self, node: Node, service: Service) -> None:
        """
        Set given service as a shell. Unlike the method add_service, set_shell does not produce undefined results, when
        a service already present at the node is set as a shell.

        :param node: An instance of a node to set shell at.
        :type node: Node

        :param service: An instance of a service to set.
        :type service: Service

        :return: None
        """

    @abstractmethod
    def add_traffic_processor(self, node: Node, processor: ActiveService) -> None:
        """
        Add a service to a traffic processor queue. Traffic processors process messages incoming into a node before any
        evaluation takes place. There can be multiple traffic processors on one node and their order depends on the
        order of add_traffic_processor calls. The first one added, processes the traffic first, and so on...

        Any active service can be set as a traffic processor. Traffic processors are free to change the contents of
        messages and unless they return False from their process_message() function that message is passed to next
        processors or to the node.

        :param node: An instance of the node where traffic processor should be added.
        :type node: Node

        :param processor: An instance of an active service which will act as a traffic processor.
        :type processor: ActiveService

        :return: None
        """

    @abstractmethod
    def add_route(self, node: Node, *route: Route) -> None:
        """
        Add a network route to a router.

        :param node: An instance of the router. Although the type is set as a Node here, this function can only be
            called on routers. Trying to assign a route to an ordinary node would do nothing.
        :type node: Node

        :param route: On or more network routes to add.
        :type route: Route

        :return: None
        """

    @abstractmethod
    def add_routing_rule(self, node: Node, rule: FirewallRule) -> None:
        """
        Adds a rule for source-destination routing. This effectively bypasses the routing rules and enables custom
        port forwarding.

        TODO
            This is only temporary - first, it is leaking implementation detail to outside and second, it is
            completely stupid, as router should be a designated active service and should provide configuration
            interface.

        :param node: An instance of a router. Although the type is set as a Node here, this function can only be
            called on routers. Trying to add a routing rule to an ordinary node would do nothing.
        :type node: Node

        :param rule: A firewall rule, which coincidentally has all the necessary ingredients for implementation of
            source-destination routing. See the TODO on why this is a stupid idea.
        :type rule: FirewallRule

        :return: None
        """

    @abstractmethod
    def set_routing_policy(self, node: Node, policy: FirewallPolicy) -> None:
        """
        Sets a default routing policy for source-destination routing.

        The same comment regarding the stupidity of this interface applies as an function add_routing_rule().

        :param node: An instance of a router. Although the type is set as a Node here, this function can only be
            called on routers. Trying to assign a routing policy to an ordinary node would do nothing.

        :param policy: A firewall policy, which coincidentally has all the necessary ingredients for implementation of
            source-destination routing policy. See the add_routing_rule() on why this is a stupid idea.
        :type policy: FirewallPolicy

        :return: None
        """

    @abstractmethod
    def list_routes(self, node: Node) -> List[Route]:
        """
        List all the routes currently set at a router.

        TODO
            This will be changed in the future, after router is implemented as an active service with its own interface.

        :param node: The router to get routes from. Although the type is set as a Node here, this function can only be
            called on routers. Trying to list routes from an ordinary node would do nothing.
        :type node: Node

        :return: A list of routes.
        """


class ServiceConfiguration(ABC):
    """
    Service configuration enables management of passive and active services.
    """
    @abstractmethod
    def create_active_service(self, type: str, owner: str, name: str, node: Node,
                              service_access_level: AccessLevel = AccessLevel.LIMITED,
                              configuration: Optional[Dict[str, Any]] = None, id: str = "") -> Optional[Service]:
        """
        Creates an active service. These include anything from traffic processors to agents.

        :param type: A type of the service, which is a unique identification among services.
        :type type: str

        :param id: An ID that will be given to the service and which must be unique within the simulation environment.
        :type id: str

        :param owner: An identity that will own the service at the node. While it should be an identity that is present
            at the node, this is currently not controlled in any way, because the whole identity business is not yet
            sorted out.
        :type owner: str

        :param name: The name of the service as it was registered into the system. (see cyst.api.host.service.ActiveServiceDescription)
        :type name:  str

        :param node: An instance of a node where the service will reside in the end. There is a technical limitation
            that lies behind this requirement. If it weren't there, everything would be done in the add_service() call.
        :type node: Node

        :param service_access_level: An access level for the service. This mostly concerns the resources of the Node it
            can access.
        :type service_access_level: AccessLevel

        :param configuration: A dictionary of configuration parameters. There are no limitations to the param/value
            pairs, as they are directly passed to function creation function. (see cyst.api.host.service.ActiveServiceDescription)
        :type configuration: Optional[Dict[str, Any]]

        :param id: An ID that will be given to the service and which must be unique within the simulation environment.
                   If no ID is provided a unique one is generated.
        :type id: str

        :return: An instance of an active service, or null if failed to create one.
        """

    @abstractmethod
    def get_service_interface(self, service: ActiveService, control_interface_type: Type[ActiveServiceInterfaceType]) -> ActiveServiceInterfaceType:
        """
        Active services can provide an interface, which enables a finer control over their function than a rather
        limited cyst.api.host.service.ActiveService interface.

        Due to python's handling of types and interfaces, you can always just call the functions of the service
        directly, without going through the hassle of interface acquisition, but such approach would not pass through
        a static verification.

        Example:

        .. code-block:: python

            from cyst.api.environment.environment import Environment
            from cyst.services.scripted_attacker.main import ScriptedAttackerControl

            env = Environment.create()
            node = env.configuration.node.create_node("node1")
            service = env.configuration.service.create_active_service("attacker1", "attacker", "scripted_attacker", node)
            ctrl = env.configuration.service.get_service_interface(service.active_service, ScriptedAttackerControl)

            ctrl.execute_action(...)

        :param service: An instance of an active service. Please consult the example and the Service interface to
            prevent unwanted surprises (.active_service).
        :type service: ActiveService

        :param control_interface_type: A type of the interface that should be obtained from the service. Note that a
            service can provide multiple interfaces.
        :type control_interface_type: Type[TypeVar('ActiveServiceInterfaceType')]

        :return: An interface of a service.
        """

    @abstractmethod
    def create_passive_service(self, type: str, owner: str, version: str = "0.0.0", local: bool = False,
                               service_access_level: AccessLevel = AccessLevel.LIMITED, id: str = "") -> Service:
        """
        Create a passive service.

        :param type: A type of the service, such as lighttpd or bash. The type is relevant for mapping exploits.
        :type type: str

        :param owner: An identity that will own the service at the node. While it should be an identity that is present
            at the node, this is currently not controlled in any way, because the whole identity business is not yet
            sorted out.
        :type owner: str

        :param version: The version of the service. The format of the version should be compatible with semantic
            versioning scheme. TODO: Extend the signature to Union[str, VersionInfo]
        :type version: str

        :param local: An indication, whether the service is accessible from outside the node.
        :type local: bool

        :param service_access_level: An access level of the service. This mostly concerns the resources than can be
            accessed in case of service compromise.
        :type service_access_level: AccessLevel

        :param id: A unique ID within the simulation. An ID is generated if not provided.
        :type id: str

        :return: An instance of passive service.
        """

    @abstractmethod
    def update_service_version(self, service: PassiveService, version: str = "0.0.0") -> None:
        """
        Changes the service version to a new value. This function is primarily meant as a way to simulate updates
        or patching of already running services.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :param version: A new value of the version. The format of the version should be compatible with semantic
            versioning scheme. TODO: Extend the signature to Union[str, VersionInfo]
        :type version: str

        :return: None
        """

    @abstractmethod
    def set_service_parameter(self, service: PassiveService, parameter: ServiceParameter, value: Any) -> None:
        """
        Sets a parameter of an existing passive service. The parameters are chosen from a parametrization domain
        defined by the ServiceParameter flags.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :param parameter: A type of the parameter to set.
        :type parameter: ServiceParameter

        :param value: A value of the parameter. For allowed values, consult cyst.api.environment.configuration.ServiceParameter.
        :type value: Any

        :return: None
        """

    @abstractmethod
    def create_data(self, id: Optional[str], owner: str, path: str, description: str) -> Data:
        """
        Creates a data. Currently, this is a very rudimentary approach and future updates are expected to provide more
        sophisticated mechanisms for data description and manipulation.

        :param id: An optional identification of the data.
        :type id: Optional[str]

        :param owner: An identity of the owner of the data. This should be an identity existing within the simulation
            (not necessarily on the node and service where the data will reside), as the access to the data should be
            checked against the name in the authentication token, which has to be the same as the owner. [That "should"
            here means that it is dependent on the implementation of behavioral models.]
        :type owner: str

        :param path: A path to the data, which should be unique among all data within one service.
        :type path: str

        :param description: A textual description of the data, which currently enables differentiation between data
            instances.
        :type description: str

        :return: An instance of data.
        """

    @abstractmethod
    def public_data(self, service: PassiveService) -> List[Data]:
        """
        Gets access to the data that should be available on the service without the need for authentication. This can
        be used both to get and set the data.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :return: A list of public data instances on the service.
        """

    @abstractmethod
    def private_data(self, service: PassiveService) -> List[Data]:
        """
        Gets access to the data that should be available on the service only after authentication. This can
        be used both to get and set the data.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :return: A list of private data instances on the service.
        """

    @abstractmethod
    def public_authorizations(self, service: PassiveService) -> List[Authorization]:
        """
        Gets access to public authorizations present at the service.

        Warning:
            This function is deprecated and will be removed soon, as it does not conform to the current
            authentication/authorization framework.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :return: A list of publicly available authorization instances.
        """

    @abstractmethod
    def private_authorizations(self, service: PassiveService) -> List[Authorization]:
        """
        Gets access to private authorizations present at the service.

        Warning:
            This function is deprecated and will be removed soon, as it does not conform to the current
            authentication/authorization framework.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :return: A list of privately available authorization instances.
        """

    @abstractmethod
    def sessions(self, service: PassiveService) -> List[Session]:
        """
        Gets access to sessions that are connected to the service. These sessions can either be the ones created as
        a part of the initialization process, or those that were created following the activities in the simulation.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :return: A list of sessions connected to the service.
        """

    @abstractmethod
    def provides_auth(self, service: Service, auth_provider: AuthenticationProvider) -> None:
        """
        Adds an authentication provider to a service.

        TODO
            This function, if it is really intended to stay, must be renamed to something better, such as add_auth_provider.
            Also the parameter type should be checked, because the implementation conflates Service and PassiveService.

        :param service: An instance of the passive service.
        :type service: Service (but should be PassiveService)

        :param auth_provider: An authentication provider to add.
        :type auth_provider: AuthenticationProvider

        :return: None
        """

    @abstractmethod
    def set_scheme(self, service: PassiveService, scheme: AccessScheme) -> None:
        """
        Sets and authentication scheme at the service. More on authentication schemes at the user's documentation or in
        the documentation of cyst.api.logic.access.

        :param service: An instance of the passive service.
        :type service: PassiveService

        :param scheme: An instance of the access scheme.
        :type scheme: AccessScheme

        :return: None
        """


class NetworkConfiguration(ABC):
    """
    Network configuration enables placing nodes inside the topology and manipulation with connection and sessions.
    Currently, it is mostly additive, but it will include all manipulation options in the future.
    """
    @abstractmethod
    def add_node(self, node: Node) -> None:
        """
        Adds a node into the topology. After a node is created, it must be added to the simulation topology. Otherwise
        it will not be included in messaging. Note that this is done implicitly when using declarative configuration,
        so this mostly concerns scenarios, such as honeypot deployment.

        :param node: An instance of the node.
        :type node: Node

        :return: None
        """

    @abstractmethod
    def add_connection(self, source: Node, target: Node, source_port_index: int = -1, target_port_index: int = -1,
                       net: str = "", connection: Optional[Connection] = None) -> Connection:
        """
        Adds a connection between two nodes, provided each one has a network port available. If one of the nodes is
        router, it may trigger DHCP address assignment, depending on the settings of node interfaces.

        :param source: An instance of the source node. Connections are currently bi-directional, so it does not
            matter which one is the source and which one is the destination.
        :type source: Node

        :param target: An instance of the destination node.
        :type target: Node

        :param source_port_index: An index of the interface at the source node. If the index does not point to an
            existing interface, an exception will be thrown.
        :type source_port_index: int

        :param target_port_index: An index of the interface at the destination node. If the index does not point to an
            existing interface, an exception will be thrown.
        :type target_port_index: int

        :param net: A network mask specification, used by a router (if one is present) to determine a correct address
            range. This parameter is not necessary, if all addresses are fully specified in the interfaces.
        :type net: str

        :param connection: An instance of a connection. This parameter enables passing a specific connection instance,
            which has, e.g., specific connection properties. If this parameter is omitted, then a new connection is
            automatically created.
        :type connection: Connection

        :return: An instance of a connection that was established between selected nodes.
        """

    @abstractmethod
    def get_connections(self, node: Node, port_index: Optional[int] = None) -> List[Connection]:
        """
        Gets connections that are connected to the node at the port index, if specified.

        :param node: An instance of the node.
        :type node: Node

        :param port_index: An index of the port.
        :type port_index: Optional[int]

        :return: A list of connections that are connected to the node.
        """

    @abstractmethod
    def create_session(self, owner: str, waypoints: List[Union[str, Node]], src_service: Optional[str] = None,
                       dst_service: Optional[str] = None, parent: Optional[Session] = None, defer: bool = False,
                       reverse: bool = False, id: Optional[str] = None) -> Optional[Session]:
        """
        Creates a fixed session in the simulated infrastructure. This session ignores the routing limitations imposed
        by router configuration. However, the creation mechanism checks if there exists a connection between each of
        session waypoints and it selects the appropriate interfaces to realize the connection through.

        :param owner: The owner of the session. This parameter is used to prevent sharing of sessions, but it will
            likely be deprecated in future versions, because it does not work well when appending sessions through
            exploiting.
        :type owner: str

        :param waypoints: A list of node instances or their IDs, which constitute the path of the session. As per
            the description, there has to be a connection between all subsequent tuples of waypoints.
        :type waypoints: List[Union[str, Node]]

        :param src_service: An ID of the service that is the anchor at the origin of the session. This service can
                            tear down the session at will.
        :type src_service: Optional[str]

        :param dst_service: An ID of the service that is the anchor at the destination of the session. This service can
                            tear down the session at will.
        :type dst_service: Optional[str]

        :param parent: A parent session that can be optionally used as a first part/head of the session.
        :type parent: Optional[Session]

        :param defer: I this parameter is set to False, the system attempts to create the session immediately. If set to
            False, the creation is deferred to the environment init phase, when all the waypoints should already be
            instantiated.
        :type defer: bool

        :param reverse: Construct the session in the reverse order, i.e., as if it was constructed as a reverse shell.
        :type reverse: bool

        :param id: An explicit ID of the session. If none is provided, it is autogenerated.
        :type id: str

        :return: An instance of a created session (unless deferred, than it always return null)
        """

    @abstractmethod
    def append_session(self, original_session: Session, appended_session: Session) -> Session:
        """
        Creates a new session by appending a two sessions together.

        :param original_session: The session that will serve as a beginning of the session. At its endpoint the
            second session will be appended.
        :type original_session: Session

        :param appended_session: The session that will serve as the end of the session. It will be connected to the
            original session with its startpoint.
        :type appended_session: Session

        :return: A new session that is a combination of both the original ones.
        """

    @abstractmethod
    def create_session_from_message(self, message: Message, reverse_direction: bool = False) -> Session:
        """
        Establishes a new session from the path the message travelled. This function is mostly used by behavioral models
        to create new sessions in response to messages coming from attackers.

        :param message: An instance of the message.
        :type message: Message
        :param reverse_direction: Whether the direction of the shell is reversed or not.
        :type reverse_direction: bool

        :return: A new session.
        """


class ExploitConfiguration(ABC):
    """
    This interface provides means to manipulate with exploits.
    """
    @abstractmethod
    def create_vulnerable_service(self, id: str, min_version: str = "0.0.0", max_version: str = "0.0.0") -> VulnerableService:
        """
        Create a definition of a vulnerable service, that is used for exploit specification

        :param name: A name of the service. It is an equivalent of PassiveService/PassiveServiceConfig name.
        :type name: str

        :param min_version: A minimum version of the service that is vulnerable. Even though the type supports any string,
            using anything else than string representation of a semantic version will lead to problems.
        :type min_version: str

        :param max_version: A maximum version of the service that is vulnerable. Even though the type supports any string,
            using anything else than string representation of a semantic version will lead to problems.
        :type max_version: str

        :return: A vulnerable service description.
        """

    @abstractmethod
    def create_exploit_parameter(self, exploit_type: ExploitParameterType, value: str = "", immutable: bool = False) -> ExploitParameter:
        """
        Creates an exploit parameter.

        Exploit parameter either represents a specification of an exploit, which can't be modified by a user, but is
        taken into account by exploit evaluation mechanisms (e.g., if the impact is one user or all users), or provides a
        mean for a user to supply additional information necessary for full execution of the exploit (e.g., identity of
        a user to impersonate).

        :param type: A type of the exploit parameter.
        :type type: ExploitParameterType

        :param value: A value of the parameter.
        :type value: Optional[str]

        :param immutable: A flag indicating, whether a user can change the value of the parameter.
        :type immutable: bool

        :return: An instance of exploit parameter.
        """

    @abstractmethod
    def create_exploit(self, id: str = "", services: List[VulnerableService] = None, locality:
                       ExploitLocality = ExploitLocality.NONE, category: ExploitCategory = ExploitCategory.NONE,
                       *parameters: ExploitParameter) -> Exploit:
        """
        Creates a working exploit.
        
        If a definition of exploit exists, then a services it refers to are vulnerable, as long as their version match.
        Patching the service and bumping a version over the max vulnerable version is the only countermeasure.

        :param id: A unique identification of the exploit.
        :type id: str 
        
        :param services: A list of services that are vulnerable to this exploit.
        :type services: List[VulnerableService]
    
        :param locality: Determines if the exploit can be used locally or remotely.
        :type locality: ExploitLocality
    
        :param category: Determines the category of the exploit.
        :type category: ExploitCategory
    
        :param parameters: An optional list of exploit parameters.
        :type parameters: Optional[List[ExploitParameter]]
        
        :return: An instance of the exploit.
        """

    @abstractmethod
    def add_exploit(self, *exploits: Exploit) -> None:
        """
        Adds an exploit to the exploit store, where it can be accessed by the services. Unless an exploit is added to
        the store, it cannot be used, even though it was created.

        :param exploits: One or more exploit instances to be added.
        :type exploits: Exploit

        :return: None
        """

    @abstractmethod
    def clear_exploits(self) -> None:
        """
        Removes all exploits from exploit store.

        :return: None
        """


class ActionConfiguration(ABC):
    """
    Action configuration enables creation of action parameter domains. While the actions are fully declarative in their
    description and their semantics stem from the interpretation given by behavioral models, action parameters enable
    fine-tuning of actions, that is accessible to automatic tools in a uniform manner.
    """
    @abstractmethod
    def create_action_parameter_domain_any(self) -> ActionParameterDomain:
        """
        Create an action parameter domain that can accept any parameter. This domain should be avoided at all costs,
        because it does not provide anything for automated tools to work with.

        :return: An unbounded parameter domain.
        """

    @abstractmethod
    def create_action_parameter_domain_range(self, default: int, min: int, max: int, step: int = 1) -> ActionParameterDomain:
        """
        Creates a parameter domain that is represented by a range of numbers. Ideal for expressing things like port
        ranges.

        :param default: A default value of the domain.
        :type default: int

        :param min: A minimal range value.
        :type min: int

        :param max: A maximal range value.
        :type max: int

        :param step: Enables to include only every step-th number between min and max.
        :type step: int

        :return: A parameter domain consisting of range of numbers.
        """

    @abstractmethod
    def create_action_parameter_domain_options(self, default: Any, options: List[Any]) -> ActionParameterDomain:
        """
        Creates a parameter domain that is represented by a set of possible values. The values are not limited in any
        way, but this enables automated tools to select from them.

        :param default: A default value of the domain.
        :type default: Any

        :param options: A list of values, which constitute the domain.
        :type options: List[Any]

        :return: A parameter domain consisting of set of values.
        """


class AccessConfiguration(ABC):
    """
    Access configuration enables manipulation of authentication and authorization primitives, creation of access schemes
    and evaluation of authentication tokens.
    """
    @abstractmethod
    def create_authentication_provider(self, provider_type: AuthenticationProviderType,
                                       token_type: AuthenticationTokenType, security: AuthenticationTokenSecurity,
                                       ip: Optional[IPAddress], timeout: int, id: str = "") -> AuthenticationProvider:
        """
        Authentication provider represents an authentication mechanism that can be employed in services via the access
        scheme mechanism.

        :param provider_type: The type of authentication provider.
        :type provider_type: AuthenticationProviderType

        :param token_type: The type of tokens that are employed by this authentication provider.
        :type token_type: AuthenticationTokenType

        :param token_security: Security mechanism applied to stored tokens.
        :type token_security: AuthenticationTokenSecurity

        :param ip: An optional IP address, which is intended for remote or federated providers. It represents an IP
            address where this provider can be accessed.
        :type ip: Optional[IPAddress]

        :param timeout: A number of simulated time units that can elapse from the initiation of authentication exchange
            before the attempt is discarded as unsuccessful.
        :type timeout: int

        :param id: A unique identification of the provider within the simulation. If no value is provided, the system
                   will generate a unique one.
        :type id: str

        :return: An authentication provider instance.
        """

    @abstractmethod
    def create_authentication_token(self, type: AuthenticationTokenType, security: AuthenticationTokenSecurity,
                                    identity: str, is_local: bool) -> AuthenticationToken:
        """
        Creates an authentication token for a given identity. This token is given a unique identifier, which
        distinguishes it from other tokens.

        Warning:
            The documentation is missing the description of the is_local parameter and it implications

        :param type: A type of the authentication token.
        :type type: AuthenticationTokenType

        :param security: A level of security for the authentication token.
        :type security: AuthenticationTokenSecurity

        :param identity: A identity this token is bound to. In theory, no identity needs to be provided (such as
            access code to doors). However, not using it here may have unintended consequences.

        :param is_local: No idea...
        :type is_local: bool

        :return: An instance of authentication token.
        """

    @abstractmethod
    def register_authentication_token(self, provider: AuthenticationProvider, token: AuthenticationToken) -> bool:
        """
        Registers an authentication token with the authentication provider. After that, anyone possessing the token
        can authenticate against the provider.

        :param provider: An instance of the authentication provider.
        :type provider: AuthenticationProvider

        :param token: An instance of the authentication token.
        :type token: AuthenticationToken

        :return: Indication, whether registration was successful.
        """

    @abstractmethod
    def unregister_authentication_token(self, token_identity: str, provider: AuthenticationProvider) -> None:
        """
        Unregisters authentication token from given provider based on given token identity.

        :param token_identity: An identity for token to be unregistered from provider.
        :type token_identity: str

        :param provider: An authentication provider to unregister token from.
        :type provider: AuthenticationProvider

        :return: None
        """

    @abstractmethod
    def create_and_register_authentication_token(self, provider: AuthenticationProvider, identity: str) -> Optional[AuthenticationToken]:
        """
        A convenience function that combines the effects of create_authentication_token() and
        register_authentication_token().

        :param provider: An instance of the authentication provider.
        :type provider: AuthenticationProvider

        :param identity: A identity the created token is bound to. In theory, no identity needs to be provided (such as
            access code to doors). However, not using it here may have unintended consequences.
        :type identity: str

        :return: An instance of authentication token that is already registered.
        """

    @abstractmethod
    def create_authorization(self, identity: str, access_level: AccessLevel, id: str, nodes: Optional[List[str]] = None,
                             services: Optional[List[str]] = None) -> Authorization:
        """
        Creates and authorization for a given identity. The authorization gives the identity access to selected nodes
        and services with a given access level.

        :param identity: An identity for whom the authorization is created.
        :type identity: str

        :param access_level: An access level this authorization provides.
        :type access_level: AccessLevel

        :param id: A unique identifier of this authorization. TODO: I have no idea, what is the purpose.
        :type id: str

        :param nodes: A list of node ids this authorization works on. This is intended for federated identities, which
            can are shared among different nodes and services. For local authentication providers, this parameters
            is of no use.
        :type nodes: Optional[List[str]]

        :param services: A list of service ids this authorization works on. This is intended for federated identities,
            which can are shared among different nodes and services. For local authentication providers, this parameters
            is of no use.
        :type services: Optional[List[str]]

        :return: An authorization token.
        """

    @abstractmethod
    def create_access_scheme(self, id: str = "") -> AccessScheme:
        """
        Creates an empty access scheme. The access scheme is a combination of authentication providers, which use a
        supplied authorizations. The access scheme provides means to describe multiple authentication schemes
        within one service or multi-factor authentication.

        :param id: A unique identification of the scheme within the simulation. If no value is provided, the system
                   will generate a unique one.
        :type id: str

        :return: An empty access scheme.
        """

    @abstractmethod
    def add_provider_to_scheme(self, provider : AuthenticationProvider, scheme: AccessScheme) -> None:
        """
        Adds an authentication provider to the access scheme.

        :param provider: An authentication provider.
        :type provider: AuthenticationProvider

        :param scheme: An access scheme to add provider to.
        :type scheme: AccessScheme

        :return: None
        """

    @abstractmethod
    def add_authorization_to_scheme(self, auth: Authorization, scheme: AccessScheme) -> None:
        """
        Adds an authorization to the access scheme.

        :param auth: An instance of authorization.
        :type auth: Authorization

        :param scheme: An access scheme to add authorization to.
        :type scheme: AccessScheme

        :return: None
        """

    @abstractmethod
    def remove_authorization_from_scheme(self, auth: Authorization, scheme: AccessScheme) -> None:
        """
        Removes given authorization from given scheme by setting its access level to AccessLevel.NONE

        :param auth: An authorization to be removed from scheme.
        :type auth: Authorization

        :param scheme: A scheme.
        :type scheme: AccessScheme

        :return: None
        """

    @abstractmethod
    def evaluate_token_for_service(self, service: Service, token: AuthenticationToken, node: Node,
                                   fallback_ip: Optional[IPAddress])\
            -> Optional[Union[Authorization, AuthenticationTarget]]:
        """
        Evaluates if a token authenticates against a service. For single-factor authentications, this function returns
        either a working authorization token or none, depending on success. For multi-factor authentications, if it is
        the last factor, then it behaves as a single-factor authentication. If it is not, then it will either return a
        next authentication target in given scheme or none, depending on success.

        TODO
            Add the description of fallback_ip.

        :param service: An instance of the service the evaluation is against.
        :type service: Service

        :param token: An instance of the authentication token.
        :type token: AuthenticationToken

        :param node: An instance of the node where the service resides on.
        :type node: Node

        :param fallback_ip: No idea
        :type fallback_ip: Optional[IPAddress]

        :return: Either a working authorization, a next target, or just none if the authentication was not successful.
        """

    @abstractmethod
    def disable_authentication_token(self, provider: AuthenticationProvider, token: AuthenticationToken, time: int) -> None:
        """
        Disables given token until given simulation time.

        :param provider: An authentication provider that contains token to be disabled.
        :type provider: AuthenticationProvider

        :param token: An authentication token to be disabled.
        :type token: AuthenticationToken

        :param time: Simulation time
        :type time: int

        :return: None
        """

    @abstractmethod
    def enable_authentication_token(self, provider: AuthenticationProvider, token: AuthenticationToken) -> None:
        """
        Enables given token.

        :param provider: An authentication provider that contains token to be enabled.
        :type provider: AuthenticationProvider

        :param token: An authentication token to be enabled.
        :type token: AuthenticationToken

        :return: None
        """

    @abstractmethod
    def create_service_access(self, service: Service, identity: str, access_level: AccessLevel,
                              tokens: List[AuthenticationToken] = None) -> Optional[List[AuthenticationToken]]:
        """
        Create a means to access the given passive service by registering authentication tokens
        under the given identity.

        TODO: Currently works only for local providers.

        :param service: A passive service to grant access to.
        :type service: PassiveService

        :param identity: An identity of the user which will gain the access.
                         It must not be already present in one of the access schemes.
        :type identity: str

        :param access_level: The level of access which will be created.
        :type access_level: AccessLevel

        :param tokens: Tokens to be optionally used instead of creating new ones.
        :type tokens: List[AuthenticationToken]

        :return: List of authentication tokens if successful, None otherwise.
        """

    @abstractmethod
    def modify_existing_access(self, service: Service, identity: str, access_level: AccessLevel) -> bool:
        """
        Modify the access level of the given passive service. No new tokens are created.

        TODO: Currently works only for local providers.

        :param service: A passive service to grant access to.
        :type service: PassiveService

        :param identity: An identity of the user which access will be modified.
                         It must be already present at one of the access scheme.
        :type identity: str

        :param access_level: The level of access which will be modified.
        :type access_level: AccessLevel

        :return: True if successful, False otherwise.
        """

class PhysicalConfiguration(ABC):
    """
    Physical configuration enables manipulation with physical locations, their assets, and users within them.
    """

    @abstractmethod
    def create_physical_location(self, location_id: str | None) -> PhysicalLocation:
        """
        Creates a new physical location. This location is empty an inaccessible.

        :param location_id: An identification of a location. If not provided, the id is autogenerated.
        :type location_id: str | None

        :return: A physical location handle.
        """

    @abstractmethod
    def get_physical_location(self, location_id: str) -> PhysicalLocation | None:
        """
        Returns a physical location handle by given id.

        :param location_id: An identification of a location.
        :type location_id: str

        :return: A physical location handle if it exists with the given id, or None otherwise.
        """

    @abstractmethod
    def get_physical_locations(self) -> List[PhysicalLocation]:
        """
        Returns all physical locations that are present in the current run.

        :return: List of physical locations.
        """

    @abstractmethod
    def remove_physical_location(self, location_id: str) -> None:
        """
        Removes a physical locations. Physical connections associated with it are removed as well. If there are assets
        and users at the physical location, there are moved to a limbo with other unassigned assets.

        :param location_id: An id of the location.
        :type location_id: str

        :return: None. Throws when attempting to remove nonexistent location.
        """

    @abstractmethod
    def create_physical_access(self, identity: str, time_from: datetime | None, time_to: datetime | None) -> PhysicalAccess:
        """
        Creates a physical access specification. This specification can be used multiple times for different locations.

        :param identity: The identity of the user for which the access is being made. The identity is not checked for
            existence, therefore, access for non-existent users can be made.
        :type identity: str

        :param time_from: A specification of a time of a day (the date part is ignored) from which the access is granted.
            If no time is provided, then 00:00:00 is assumed.
        :type time_from: datetime | None

        :param time_to: A specification of a time of a day (the date part is ignored) to which the access is granted.
            If no time is provided, then 23:59:59 is assumed.
        :type time_to: datetime | None

        :return: A physical access handle.
        """

    @abstractmethod
    def add_physical_access(self, location_id: str, access: PhysicalAccess) -> None:
        """
        Adds physical access to a given location. Accesses may overlap, asi it does not matter, because they all
        represent an ALLOW rule.

        :param location_id: The ID of a location where to add the access. Throws if nonexistent.
        :type location_id: str

        :param access: A physical access handle.
        :type access: PhysicalAccess

        :return: None
        """

    @abstractmethod
    def get_physical_accesses(self, location_id: str) -> List[PhysicalAccess]:
        """
        Lists all accesses that are in effect at a given location.

        :param location_id: The ID of a location to list accesses from. Throws if nonexistent.
        :type location_id: str

        :return: Physical access handles for the given location.
        """
        pass

    @abstractmethod
    def remove_physical_access(self, location_id: str, access: PhysicalAccess) -> None:
        """
        Removes physical access from the given location.

        :param location_id: The ID of a location to remove accesses from. Throws if nonexistent.
        :type location_id: str

        :param access: A physical access handle.
        :type access: PhysicalAccess

        :return: None
        """

    @abstractmethod
    def add_physical_connection(self, origin: str, destination: str, travel_time: Duration) -> None:
        """
        Creates a new physical connection between two locations. If such connection already exists, a new connection
        is not created.

        :param origin: An id of a location which is the first endpoint of the connection. Throws if nonexistent.
        :type origin: str

        :param destination: An id of a location which is the second endpoint of the connection. Throws if nonexistent.
        :type destination: str

        :param travel_time: A duration needed to traverse the connection.
        :type travel_time: Duration

        :return: None
        """

    @abstractmethod
    def remove_physical_connection(self, origin: str, destination: str) -> None:
        """
        Removes a physical connection between two locations. If the connection does not exist, this is no-op.

        :param origin: An id of a location which is the first endpoint of the connection. Throws if nonexistent.
        :type origin: str

        :param destination: An id of a location which is the second endpoint of the connection. Throws if nonexistent.
        :type destination: str

        :return: None
        """

    @abstractmethod
    def get_physical_connections(self, origin: str, destination: str | None) -> List[PhysicalConnection]:
        """
        Get a list of physical connections between locations.

        :param origin: An id of a location which is one of the endpoints of the physical connection. Throws if
            nonexistent.
        :type origin: str

        :param destination: An id of a location which is another one of the endpoints of the physical connections. If
            no destination is specified, then this function returns all physical connections from/to an origin. Throws
            is nonexistent.
        :type destination: str

        :return: A list of physical connections.
        """

    @abstractmethod
    def place_asset(self, location_id: str, asset: str) -> None:
        """
        Add an asset to a location. For this purpose, users are treated as assets, much like nodes and others. This is
        a deus ex machina function. The asset appears in the location instantaneously. The function only controls if the
        asset is already present elsewhere. In such case the asset is removed from the other location and placed in this
        one. If the asset is already assigned to the given location, nothing happens.

        :param location_id: An ID of a location where to place the asset. Throws if nonexistent.
        :type location_id: str

        :param asset: An ID of the asset to add to a location.
        :type asset: str

        :return: None
        """

    @abstractmethod
    def remove_asset(self, location_id: str, asset: str) -> None:
        """
        Removes an asset from a location. For this purpose, users are treated as assets, much like nodes and others.
        This is a deus ex machina function. The asset is removed from the location instantaneously and is placed in a
        limbo.

        :param location_id: An ID of a location from which to remove the asset. Throws if nonexistent.
        :type location_id: str

        :param asset: An ID of the asset to remove from a location.
        :type asset: str

        :return: None
        """

    @abstractmethod
    def move_asset(self, origin: str, destination: str, asset: str) -> Tuple[bool, str, str]:
        """
        Moves an asset between two locations. For this purpose, users are treated as assets, much like nodes and others.
        Before the move, it is checked, whether there is a physical connection between the locations. If so, then the
        asset traverses the connection for given duration, and if it is the User, it is checked, whether they have
        access rights.

        All assets are traversing the connection for the same time, so for example, if you want to mimic a moving of
        a piece of hardware between two locations, do not forget to add reasonable delays before and after the move to
        simulate packing and unpacking.

        :param origin: An id of a location which is the first endpoint of the connection. Throws if nonexistent.
        :type origin: str

        :param destination: An id of a location which is the second endpoint of the connection. Throws if nonexistent.
        :type destination: str

        :param asset: An ID of the asset to move between locations.
        :type asset: str

        :return: A tuple representing (<was the move successful?>, <what is the location of asset after the move>, <a problem description if there was any>)
        """

    @abstractmethod
    def get_assets(self, location_id: str) -> List[str]:
        """
        Get the list of assets present at the specified location.

        :param location_id: A name of a location from which to get the assets. Throws if nonexistent.
        :type location_id: str

        :return: A list of assets.
        """

    @abstractmethod
    def get_location(self, asset: str) -> str:
        """
        Get the location of a given asset.

        :param asset: The ID of the asset for which you want the location.
        :type asset: str

        :return: The id of a location, or an empty string if the asset is in limbo.
        """


# ----------------------------------------------------------------------------------------------------------------------
# Runtime configuration of the environment. Can be filled from different sources
@dataclass
class RuntimeConfiguration:
    """
    Runtime configuration represents parametrization of running core instances. This configuration can come from
    multiple sources, e.g., environment variables, files, or command line parameters.

    :param data_backend: Specification of the backend for the data store. CYST by default provides two data store
        backends - 'memory' and 'sqlite'.
    :type data_backend: str

    :param data_backend_params: Parametrization of the data backend. Once something else than parameter-less memory
        backend works, it will be documented here.
    :type data_backend_params: Dict[str, str]

    :param run_id: The current run id, regardless of it was autogenerated, or provided by a user.
    :type run_id: str

    :param run_id_log_suffix: If set to true, run_id will be a part of log file names.
    :type run_id_log_suffix: bool

    :param config_id: An identifier of a configuration for the current run, if the configuration is retrieved from a
        database.
    :type config_id: str

    :param config_filename: A path to a file where a configuration for the current run is present.
    :type config_filename: str

    :param max_running_time: An upper limit on an execution time of a run. A platform time is considered, not the real
        time. The run is not guaranteed to finish at exactly the specified time, rather it will gracefully finish if the
        running time is exceeded.
    :type max_running_time: float

    :param max_action_count: An upper limit on the number of executed actions by any actor. When this count is reached,
        the run terminates. Only the top-level actions (i.e., not actions executed within composite actions) are counted
        and the run is not guaranteed to finish at exactly the specified action count. Rather, it will gracefully finish
        if the action count is exceeded.
    :type max_action_count: int

    :param other_params: A dictionary of parameters that are not directly consumed by the environment but are passed to
        the other components, agents, etc.
    :type other_params: Dict[str, str]
    """
    data_backend: str = "memory"
    data_backend_params: Dict[str, str] = field(default_factory=lambda: {})
    run_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    run_id_log_suffix: bool = False
    config_id: str = ""
    config_filename: str = ""
    max_running_time: float = 0.0
    max_action_count: int = 0
    other_params: Dict[str, str] = field(default_factory=lambda: {})



from typing import Tuple, Callable, Union, List

from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.message import Request, Response, Status, StatusOrigin, StatusValue
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.platform_specification import PlatformSpecification, PlatformType
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.logic.action import ActionDescription, ActionParameterType, ActionParameter, Action, ActionType
from cyst.api.logic.behavioral_model import BehavioralModel, BehavioralModelDescription
from cyst.api.logic.composite_action import CompositeActionManager
from cyst.api.network.node import Node
from cyst.api.utils.duration import Duration, msecs


class CYSTModel(BehavioralModel):
    def __init__(self, configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                 policy: EnvironmentPolicy, messaging: EnvironmentMessaging, infrastructure: EnvironmentInfrastructure,
                 composite_action_manager: CompositeActionManager) -> None:

        self._configuration = configuration
        self._action_store = resources.action_store
        self._exploit_store = resources.exploit_store
        self._policy = policy
        self._messaging = messaging
        self._infrastructure = infrastructure
        self._cam = composite_action_manager

        self._action_store.add(ActionDescription(id="cyst:test:echo_success",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="A testing message that returns a SERVICE|SUCCESS",
                                                 parameters=[ActionParameter(ActionParameterType.NONE, "punch_strength",
                                                                             configuration.action.create_action_parameter_domain_options("weak", ["weak", "super strong"]))]))

        self._action_store.add(ActionDescription(id="cyst:test:echo_failure",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="A testing message that returns a SERVICE|FAILURE",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:test:echo_error",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="A testing message that returns a SERVICE|ERROR",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:network:create_session",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Create a session to a destination service",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:host:get_services",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Get list of services on target node",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:host:get_remote_services",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Get list of services on target node",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:host:get_local_services",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Get list of services on target node",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:compound:session_after_exploit",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Create a session after a successful application of an exploit",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:active_service:open_session",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Open a session to an existing active service acting as forward/reverse shell.",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="cyst:active_service:action_1",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="A placeholder action for active services instead of dedicated behavioral model.",
                                                 parameters=[]))

    async def action_flow(self, message: Request) -> Tuple[Duration, Response]:
        raise RuntimeError("CYST namespace does not support composite actions")

    async def action_effect(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        if not message.action:
            raise ValueError("Action not provided")

        action_name = "_".join(message.action.fragments)
        fn: Callable[[Request, Node], Tuple[Duration, Response]] = getattr(self, "process_" + action_name, self.process_default)
        return fn(message, node)

    def action_components(self, message: Union[Request, Response]) -> List[Action]:
        # CYST actions are component-less
        return []

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:TEST
    def process_default(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        print("Could not evaluate message. Tag in `cyst` namespace unknown. " + str(message))
        return msecs(0), self._messaging.create_response(message, status=Status(StatusOrigin.SYSTEM, StatusValue.ERROR), session=message.session)

    def process_test_echo_success(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        return msecs(20), self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                  session=message.session, auth=message.auth)

    def process_test_echo_failure(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        return msecs(20), self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.FAILURE),
                                                          session=message.session, auth=message.auth)

    def process_test_echo_error(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        return msecs(20), self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.ERROR),
                                                          session=message.session, auth=message.auth)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:NETWORK
    def process_network_create_session(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        session = self._configuration.network.create_session_from_message(message)
        return msecs(40), self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                          session=session, auth=message.auth)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:HOST
    def process_host_get_services(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        services = []
        for service in node.services.values():
            if service.passive_service:
                services.append((service.name, service.passive_service.version))
        return msecs(40), self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                          session=message.session, auth=message.auth, content=services)

    def process_host_get_remote_services(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        services = []
        for service in node.services.values():
            if service.passive_service and not service.passive_service.local:
                services.append((service.name, service.passive_service.version))
        return msecs(40), self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                          session=message.session, auth=message.auth, content=services)

    def process_host_get_local_services(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        services = []
        for service in node.services.values():
            if service.passive_service and service.passive_service.local:
                services.append((service.name, service.passive_service.version))
        return msecs(40), self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                          session=message.session, auth=message.auth, content=services)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:COMPOUND
    def process_compound_session_after_exploit(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        # TODO: Add check if exploit category and locality is ok
        # Check if the service is running on the target
        error = ""
        if not message.dst_service:
            error = "Service for session creation not specified"
        # and that an exploit is provided
        elif not message.action.exploit:
            error = "Exploit not specified to ensure session creation"
        # and it actually works
        elif not self._exploit_store.evaluate_exploit(message.action.exploit, message, node):
            error = f"Service {message.dst_service} not exploitable using the exploit {message.action.exploit.id}"

        if error:
            return msecs(20), self._messaging.create_response(message, Status(StatusOrigin.NODE, StatusValue.ERROR), error,
                                                              session=message.session)
        else:
            return msecs(100), self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                               session=self._configuration.network.create_session_from_message(message),
                                                               auth=message.auth)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:ACTIVE_SERVICE
    def process_active_service_action_1(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        # These actions cannot be called on passive services
        return msecs(60), self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                          "Cannot call active service placeholder actions on passive services.",
                                                          session=message.session)

    def process_active_service_action_2(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        # These actions cannot be called on passive services
        return msecs(60), self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                          "Cannot call active service placeholder actions on passive services.",
                                                          session=message.session)

    def process_active_service_action_3(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        # These actions cannot be called on passive services
        return msecs(60), self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                          "Cannot call active service placeholder actions on passive services.",
                                                          session=message.session)

    def process_active_service_open_session(self, message: Request, node: Node):
        # These actions cannot be called on passive services
        return msecs(40), self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                          "Cannot open session with service that's not an active shell service",
                                                          session=message.session)


def create_cyst_model(configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                      policy: EnvironmentPolicy, messaging: EnvironmentMessaging,
                      infrastructure: EnvironmentInfrastructure,
                      composite_action_manager: CompositeActionManager) -> BehavioralModel:
    model = CYSTModel(configuration, resources, policy, messaging, infrastructure, composite_action_manager)
    return model


behavioral_model_description = BehavioralModelDescription(
    namespace="cyst",
    description="Behavioral model that is equivalent to CYST actionable API",
    creation_fn=create_cyst_model,
    platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")]
)



from semver import VersionInfo
from typing import List, Set, Union, Tuple, Optional, Dict

import cyst
from cyst.api.host.service import Service, ActiveService, PassiveService, ServiceState
from cyst.api.logic.access import AccessLevel, AuthenticationToken, Authorization, AuthenticationProvider, AccessScheme
from cyst.api.logic.data import Data
from cyst.api.network.node import Node
from cyst.api.network.session import Session

from cyst.platform.logic.access import AuthorizationImpl
from cyst.platform.logic.data import DataImpl


class ServiceImpl(Service):

    def __init__(self, type: str, service: Union[ActiveService, PassiveService],
                 name: str, owner: str, service_access_level: AccessLevel = AccessLevel.LIMITED, id: str = ""):
        self._type = type
        self._service = service
        self._name = name
        self._owner = owner
        self._sal = service_access_level
        self._node = None
        self._sessions: Dict[str, Session] = dict()
        self._id = id

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> str:
        return self._type

    def set_node(self, id):
        self._node = id

    @property
    def name(self) -> str:
        return self._name

    @property
    def owner(self) -> str:
        return self._owner

    @property
    def service_access_level(self) -> AccessLevel:
        return self._sal

    @property
    def passive(self) -> bool:
        return isinstance(self._service, PassiveService)

    @property
    def sessions(self) -> Dict[str, Session]:
        return self._sessions

    @sessions.setter
    def sessions(self, value: Dict[str, Session]) -> None:
        self._sessions = value

    def add_session(self, session: Session) -> None:
        if session.id not in self._sessions:
            self._sessions[session.id] = session

    @property
    def passive_service(self) -> Optional[PassiveService]:
        if isinstance(self._service, PassiveService):
            return self._service
        return None

    @property
    def active_service(self) -> Optional[ActiveService]:
        if isinstance(self._service, ActiveService):
            return self._service
        return None

    @staticmethod
    def cast_from(o: Service) -> 'ServiceImpl':
        if isinstance(o, ServiceImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Service interface")


class PassiveServiceImpl(ServiceImpl, PassiveService):
    def __init__(self, type: str, owner: str, version: str = "0.0.0", local: bool = False,
                 service_access_level: AccessLevel = AccessLevel.LIMITED, id: str = "") -> None:
        super(PassiveServiceImpl, self).__init__(type, self, type, owner, service_access_level, id)

        self._state = ServiceState.INIT
        self._version = VersionInfo.parse(version)
        self._public_data: List[Data] = []
        self._private_data: List[Data] = []
        self._public_authorizations: List[Authorization] = []
        self._private_authorizations: List[Authorization] = []
        self._tags: Set[str] = set()  # Mypy: I did not find out the type
        self._enable_session = False
        self._session_access_level = AccessLevel.NONE
        self._local = local
        self._provided_auths: List[AuthenticationProvider] = []
        self._access_schemes: List[AccessScheme] = []
        self._active_authorizations: List[Authorization] = []
        self._authentication_tokens = set()

    # ------------------------------------------------------------------------------------------------------------------
    # PassiveService
    @property
    def version(self) -> VersionInfo:
        return self._version

    @version.setter
    def version(self, value: str) -> None:
        self._version = VersionInfo.parse(value)

    @property
    def state(self) -> ServiceState:
        return self._state

    def set_state(self, value: ServiceState) -> None:
        self._state = value

    @property
    def tags(self):
        return self._tags

    @property
    def local(self) -> bool:
        return self._local

    # ------------------------------------------------------------------------------------------------------------------

    def add_public_data(self, data: DataImpl):
        self._public_data.append(data)

    def add_private_data(self, data: DataImpl):
        self._private_data.append(data)

    def add_public_authorization(self, *authorization: Authorization) -> None:
        for auth in authorization:
            self._public_authorizations.append(auth)

    def add_private_authorization(self, *authorization: Authorization) -> None:
        for auth in authorization:
            self._private_authorizations.append(auth)

    def add_tags(self, *tags):
        for tag in tags:
            self._tags.add(tag)

    def add_provider(self, provider: AuthenticationProvider):
        self._provided_auths.append(provider)
        if isinstance(provider, cyst.platform.logic.access.AuthenticationProviderImpl):
            provider.set_service(self._id)

    def add_access_scheme(self, scheme: AccessScheme):
        self._access_schemes.append(scheme)

    def add_active_authorization(self, auth: Authorization):
        self._active_authorizations.append(auth)

    def add_authentication_token(self, token: AuthenticationToken):
        self._authentication_tokens.add(token)

    def assess_authorization(self, auth: Authorization, access_level: AccessLevel, node: str,
                             service: str) -> Tuple[bool, str]:
        auth = AuthorizationImpl.cast_from(auth)

        # workaround for federated authorization
        if auth.services == ["*"] and (auth.nodes == ["*"] or node in auth.nodes):
            return True, "Federated authorization is valid"

        for active_auth in map(AuthorizationImpl.cast_from, self._active_authorizations):
            if auth.matching_id(active_auth) and access_level <= active_auth.access_level and \
                    node in active_auth.nodes and service in active_auth.services:
                return True, "Authorization is valid."
        return False, "Invalid authorization."

    @property
    def private_data(self) -> List[Data]:
        return self._private_data

    @property
    def public_data(self) -> List[Data]:
        return self._public_data

    @property
    def private_authorizations(self) -> List[Authorization]:
        return self._private_authorizations

    @property
    def public_authorizations(self) -> List[Authorization]:
        return self._public_authorizations

    @property
    def authentication_tokens(self) -> Set[AuthenticationToken]:
        return self._authentication_tokens

    @property
    def enable_session(self) -> bool:
        return self._enable_session

    @property
    def access_schemes(self) -> List[AccessScheme]:
        return self._access_schemes

    def set_enable_session(self, value: bool) -> None:
        self._enable_session = value

    @property
    def session_access_level(self) -> AccessLevel:
        return self._session_access_level

    def set_session_access_level(self, value) -> None:
        self._session_access_level = value

    @staticmethod
    def cast_from(o: PassiveService) -> 'PassiveServiceImpl':
        if isinstance(o, PassiveServiceImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the PassiveService interface")



from abc import abstractmethod, ABC
from asyncio import Task
from dataclasses import dataclass
from typing import Callable

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.environment.clock import Clock
from cyst.api.environment.configuration import EnvironmentConfiguration, GeneralConfiguration, ActionConfiguration, ExploitConfiguration, PhysicalConfiguration
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.platform_interface import PlatformInterface
from cyst.api.environment.platform_specification import PlatformSpecification
from cyst.api.environment.resources import EnvironmentResources


class Platform(ABC):
    """
    Platform interface provides an abstraction layer for different execution environments, e.g., simulation or
    emulation. When creating a new Environment a platform is instantiated and from then it takes care of message
    delivery and infrastructure configuration. As such, it is closely related to behavioral models, whose actions are
    custom tailored for a given platform.
    """

    @abstractmethod
    def init(self) -> bool:
        """
        Initialize platform. After this call, the platform should correctly process all calls. The platform need not
        check whether the init was called, the core will execute it exactly once before calling anything else.

        If the initialization was not successful, the platform should shutdown gracefully and release all resources.
        Error should be communicated via the logging interface.

        :return: True if the initialization finished successfully. False otherwise.
        """

    @abstractmethod
    def terminate(self) -> bool:
        """
        Terminate all platform operations. Within this call, all messages should be disposed of and the configured
        infrastructure should be torn down. This function will be called only once, so there is no need for checking.

        If the termination was not successful, the platform should shutdown as much as possible and indicate problems
        via the logging interface.

        :return: True if the termination finished successfully. False otherwise.
        """

    @property
    @abstractmethod
    def configuration(self) -> EnvironmentConfiguration:
        """
        The configuration interface for the entire platform. In essence, platforms may implement up-to the extent of
        CYST simulation. But it is generally understood that the configuration options for, e.g., emulation platforms
        will be much more limited and will not enable all on-the-fly modifications.

        Thus, platforms has to implement (and really implement because of usage of abstract base classes), the
        following:

        * NodeConfiguration: every method can end with exception
        * NetworkConfiguration: every method can end with exception
        * AccessConfiguration: every method can end with exception
        * ServiceConfiguration: :func:`get_service_interface` must be implemented, the rest can end with exception
        * ActionConfiguration: the one provided by the core can be used
        * ExploitConfiguration: the one provided by the core can be used
        * GeneralConfiguration:

          * :func:`get_configuration`, :func:`save_configuration`, :func:`load_configuration`, and
            :func:`get_configuration_by_id` should be proxied back to the core. :func:`get_object_by_id` must be
            implemented.

        :return: An environment configuration interface.
        """

    @abstractmethod
    def configure(self, *config_item: ConfigItem) -> 'Platform':
        """
        Configures the platform, according to provided configuration items. This function can be called repeatedly,
        however, each subsequent should replace the previous configuration. Therefore, a configuration must be done
        at once and every later change in the environment setup must be done through the
        :class:`cyst.api.environment.configuration.EnvironmentConfiguration` interface.

        :param config_item: One or more configuration items. The number of items can be arbitrary and it is not
            order-dependent.
        :type config_item: ConfigItem

        :return: The configured platform.
        """

    @property
    @abstractmethod
    def messaging(self) -> EnvironmentMessaging:
        """
        Provides access to messaging within the platform. Considering the variable nature of different platforms, they
        are expected to supply their own implementation of different messages (
        :class:`cyst.api.environment.message.Message`, :class:`cyst.api.environment.message.Request`,
        :class:`cyst.api.environment.message.Response`, :class:`cyst.api.environment.message.Timeout`). They do not have
        to implement :class:`cyst.api.environment.message.Resource`, as this is a domain of the core.

        :return: An environment messaging interface.
        """

    @abstractmethod
    async def process(self, time_advance: float) -> bool:
        """
        The main processing loop of the platform. Within this loop, the platform should process all messages that were
        enqueued through the :class:`cyst.api.environment.messaging.EnvironmentMessaging` interface.

        :param time_advance: A hint to advance the time by the specified amount of time units. The platform is free to
            ignore the parameter in real-time based systems. In simulated systems, where the time is under the control
            of the platform the time should be advanced to prevent a deadlock.
        :type time_advance: float

        :return: True if there was a message to process, False otherwise.
        """

    @property
    @abstractmethod
    def clock(self) -> Clock:
        """
        Access to a platform time management.

        :return: Clock interface.
        """


@dataclass
class PlatformDescription:
    """
    A description of a platform that is loaded as a module to CYST.

    :param specification: A platform specification that has to be unique among all loaded modules.
    :type specification: PlatformSpecification

    :param description: A textual description of the platform and its capabilities. A user-facing string.
    :type description: str

    :param creation_fn: A platform instantiation function.
    :type creation_fn: Callable[[PlatformInterface, GeneralConfiguration, EnvironmentResources, ActionConfiguration, ExploitConfiguration, EnvironmentInfrastructure], Platform]
    """
    specification: PlatformSpecification
    description: str
    creation_fn: Callable[[PlatformInterface, GeneralConfiguration, EnvironmentResources, ActionConfiguration, ExploitConfiguration, PhysicalConfiguration, EnvironmentInfrastructure], Platform]


from abc import ABC, abstractmethod
from dataclasses import dataclass
from deprecated.sphinx import versionadded, versionchanged
from enum import Enum
from typing import Callable, List, Tuple, Union

from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.platform_specification import PlatformSpecification, PlatformType
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.message import Request, Response
from cyst.api.logic.action import Action
from cyst.api.logic.composite_action import CompositeActionManager
from cyst.api.network.node import Node
from cyst.api.utils.duration import Duration


class BehavioralModel(ABC):

    @abstractmethod
    async def action_flow(self, message: Request) -> Tuple[Duration, Response]:
        pass

    @abstractmethod
    async def action_effect(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        pass

    @abstractmethod
    def action_components(self, message: Union[Request, Response]) -> List[Action]:
        pass


@versionchanged(version="0.6.0", reason="Added platform specification")
@dataclass
class BehavioralModelDescription:
    """
    TODO: Add text

    :param namespace: A namespace in which all interpreted actions belong to. An arbitrary namespace nesting is
        supported through the dot notation (e.g., "a.b.c" namespaces and actions of the form "a.b.c.action_n").
    :type namespace: str

    :param description: A textual description of the action interpreter. The description should introduce the behavioral
        model that the interpreter implements, so that the users can decide whether to use it or not.
    :type description: str

    :param creation_fn: A factory function that can create the interpreter.
    :type creation_fn: Callable[[EnvironmentConfiguration, EnvironmentResources, EnvironmentPolicy, EnvironmentMessaging, CompositeActionManager], ActionInterpreter]

    :param platform: A platform or platforms for which this behavioral model is constructed. Different modules can
        support one namespace for different platforms.
    :type platform: Union[PlatformSpecification, List[PlatformSpecification]]

    """
    namespace: str
    description: str
    creation_fn: Callable[[EnvironmentConfiguration, EnvironmentResources, EnvironmentPolicy, EnvironmentMessaging, EnvironmentInfrastructure, CompositeActionManager], BehavioralModel]
    platform: Union[PlatformSpecification, List[PlatformSpecification]] = PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST")



from sqlalchemy import create_engine, ForeignKey
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, MappedAsDataclass, Session, relationship
from typing import Dict, List

from cyst.api.environment.data_model import ActionModel
from cyst.api.environment.stats import Statistics
from cyst.api.environment.stores import DataStore, DataStoreDescription
from cyst.api.environment.message import Status, Message, Request, Response, Signal


class Base(DeclarativeBase):
    pass


class DBAction(Base):
    __tablename__ = "action"

    id: Mapped[int] = mapped_column(primary_key=True)
    message_id: Mapped[int] = mapped_column()
    run_id: Mapped[str] = mapped_column()
    action_id: Mapped[str] = mapped_column()
    caller_id: Mapped[str] = mapped_column()
    src_ip: Mapped[str] = mapped_column()
    dst_ip: Mapped[str] = mapped_column()
    dst_service: Mapped[str] = mapped_column()
    parameters: Mapped[List['DBActionParameter']] = relationship("DBActionParameter",
                                                                 back_populates="action",
                                                                 cascade="all, delete")
    status_origin: Mapped[str] = mapped_column()
    status_value: Mapped[str] = mapped_column()
    status_detail: Mapped[str] = mapped_column()
    response: Mapped[str] = mapped_column()
    session_in: Mapped[str] = mapped_column()
    session_out: Mapped[str] = mapped_column()
    auth_in: Mapped[str] = mapped_column()
    auth_out: Mapped[str] = mapped_column()


class DBActionParameter(Base):
    __tablename__ = "action_parameter"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column()
    value: Mapped[str] = mapped_column()

    action_id: Mapped[int] = mapped_column(ForeignKey("action.id"))
    action: Mapped['DBAction'] = relationship(back_populates="parameters")


# We are not keeping a full relationship between actions and messages to make it more easily digestible for user
# analyses. So the action parameters are not really present as they are mapping by message id to one specific action
# and can be extracted if needed.
#
# The SQL query to retrieve the messages together with their parameters (for CYST simulation platform) is:
#
# select message.*,
#        max(case when platform_specific.name == 'current_hop_ip' then platform_specific.value END) as current_hop_ip,
#        max(case when platform_specific.name == 'current_hop_id' then platform_specific.value END) as current_hop_id,
#        max(case when platform_specific.name == 'next_hop_ip' then platform_specific.value END) as next_hop_ip,
#        max(case when platform_specific.name == 'next_hop_id' then platform_specific.value END) as next_hop_id
# from message
# left join platform_specific on message.id = platform_specific.message_id
# group by message.id;
class DBMessage(Base):
    __tablename__ = "message"

    id: Mapped[int] = mapped_column(primary_key=True)
    message_id: Mapped[int] = mapped_column()
    type: Mapped[str] = mapped_column()
    run_id: Mapped[str] = mapped_column()
    action_id: Mapped[str] = mapped_column()
    caller_id: Mapped[str] = mapped_column()
    src_ip: Mapped[str] = mapped_column()
    dst_ip: Mapped[str] = mapped_column()
    dst_service: Mapped[str] = mapped_column()
    ttl: Mapped[int] = mapped_column()
    status_origin: Mapped[str] = mapped_column()
    status_value: Mapped[str] = mapped_column()
    status_detail: Mapped[str] = mapped_column()
    session: Mapped[str] = mapped_column()
    auth: Mapped[str] = mapped_column()
    response: Mapped[str] = mapped_column()

    platform_specific: Mapped[List['DBMessagePlatformSpecific']] = relationship("DBMessagePlatformSpecific",
                                                                                back_populates="message",
                                                                                cascade="all, delete")

class DBMessagePlatformSpecific(Base):
    __tablename__ = "platform_specific"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column()
    value: Mapped[str] = mapped_column()

    message_id: Mapped[int] = mapped_column(ForeignKey("message.id"))
    message: Mapped[DBMessage] = relationship(back_populates="platform_specific")


class DBStatistics(Base):
    __tablename__ = "statistics"

    id: Mapped[int] = mapped_column(primary_key=True)
    run_id: Mapped[str] = mapped_column()
    configuration_id: Mapped[str] = mapped_column()
    start_time_real: Mapped[float] = mapped_column()
    end_time_real: Mapped[float] = mapped_column()
    end_time_virtual: Mapped[float] = mapped_column()


class DBSignal(Base):
    __tablename__ = "signal"

    id: Mapped[int] = mapped_column(primary_key=True)
    run_id: Mapped[str] = mapped_column()
    signal_origin: Mapped[str] = mapped_column()
    state: Mapped[str] = mapped_column()
    effect_origin: Mapped[str] = mapped_column()
    effect_message: Mapped[int] = mapped_column()
    effect_description: Mapped[str] = mapped_column()

    effect_parameters: Mapped[List['DBEffectParameter']] = relationship("DBEffectParameter",
                                                                        back_populates="signal",
                                                                        cascade="all, delete")


class DBEffectParameter(Base):
    __tablename__ = "effect_parameter"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column()
    value: Mapped[str] = mapped_column()

    signal_id: Mapped[int] = mapped_column(ForeignKey("signal.id"))
    signal: Mapped[DBSignal] = relationship(back_populates="effect_parameters")


class DataStoreSQLite(DataStore):
    def __init__(self, run_id: str, params: Dict[str, str]):
        self._run_id = run_id
        self._db_path = "cyst.db"
        if "path" in params:
            self._db_path = params["path"]

        self._db = create_engine(f"sqlite+pysqlite:///{self._db_path}")
        Base.metadata.create_all(self._db)

    def add_action(self, action: ActionModel) -> None:
        with Session(self._db) as session:
            db_action = DBAction(
                message_id=action.message_id,
                run_id=self._run_id,
                action_id=action.action_id,
                caller_id=action.caller_id,
                src_ip=action.src_ip,
                dst_ip=action.dst_ip,
                dst_service=action.dst_service,
                status_origin=action.status_origin,
                status_value=action.status_value,
                status_detail=action.status_detail,
                response=action.response,
                session_in=action.session_in,
                session_out=action.session_out,
                auth_in=action.auth_in,
                auth_out=action.auth_out
            )
            session.add(db_action)
            session.flush()

            for param in action.parameters:
                p = DBActionParameter(
                    name=param.name,
                    value=param.value,
                    action_id=db_action.id
                )
                session.add(p)

            session.commit()

    def add_message(self, message: Message) -> None:
        if isinstance(message, Request):
            action_id = message.action.id
            status_origin = ""
            status_value = ""
            status_detail = ""
            response = ""
        elif isinstance(message, Response):
            action_id = message.action.id
            status_origin = str(message.status.origin)
            status_value = str(message.status.value)
            status_detail = str(message.status.detail)
            response = str(message.content)
        else:
            action_id = ""
            status_origin = ""
            status_value = ""
            status_detail = ""
            response = ""

        with Session(self._db) as session:
            db_message = DBMessage(
                message_id=message.id,
                type=str(message.type),
                run_id=self._run_id,
                action_id=action_id,
                caller_id=message.platform_specific["caller_id"],
                src_ip=str(message.src_ip),
                dst_ip=str(message.dst_ip),
                dst_service=message.dst_service,
                ttl=message.ttl,
                status_origin=status_origin,
                status_value=status_value,
                status_detail=status_detail,
                session=str(message.session.id) if message.session else "",
                auth="",
                response=response
            )
            session.add(db_message)
            session.flush()

            for k, v in message.platform_specific.items():
                if k == "caller_id":
                    continue

                session.add(DBMessagePlatformSpecific(
                    name=k,
                    value=str(v),
                    message_id=db_message.id
                ))

            session.commit()

    def add_statistics(self, statistics: Statistics) -> None:
        with Session(self._db) as session:
            session.add(DBStatistics(
                run_id=statistics.run_id,
                configuration_id=statistics.configuration_id,
                start_time_real=statistics.start_time_real,
                end_time_real=statistics.end_time_real,
                end_time_virtual=statistics.end_time_virtual
            ))
            session.commit()

    def add_signal(self, signal: Signal) -> None:
        with Session(self._db) as session:
            db_signal = DBSignal(
                run_id=self._run_id,
                signal_origin=signal.signal_origin,
                state=str(signal.state),
                effect_origin=signal.effect_origin,
                effect_message=signal.effect_message,
                effect_description=signal.effect_description,
                effect_parameters=[]
            )
            session.add(db_signal)
            session.flush()

            for k, v in signal.effect_parameters.items():
                session.add(DBEffectParameter(
                    name=k,
                    value=str(v),
                    signal_id=db_signal.id
                ))

            session.commit()

def create_data_store_sqlite(run_id: str, params: Dict[str, str]) -> DataStore:
    return DataStoreSQLite(run_id, params)


data_store_sqlite_description = DataStoreDescription(
    backend="sqlite",
    description="A SQLite-based data store.",
    creation_fn=create_data_store_sqlite
)



from __future__ import annotations

import uuid

from typing import TYPE_CHECKING, List, Union

from netaddr import IPAddress, IPNetwork

from cyst.api.environment.configuration import NodeConfiguration
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.host.service import ActiveService, Service
from cyst.api.network.elements import Route, Interface, Port
from cyst.api.network.firewall import FirewallPolicy, FirewallRule
from cyst.api.network.node import Node

from cyst.platform.host.service import ServiceImpl
from cyst.platform.network.elements import InterfaceImpl, PortImpl
from cyst.platform.network.node import NodeImpl
from cyst.platform.network.router import Router

if TYPE_CHECKING:
    from cyst.platform.main import CYSTPlatform


class NodeConfigurationImpl(NodeConfiguration):
    def __init__(self, platform: CYSTPlatform):
        self._platform = platform

    def create_node(self, id: str, ip: Union[str, IPAddress] = "", mask: str = "", shell: Service = None) -> Node:
        return _create_node(self._platform, id, ip, mask, shell)

    def create_router(self, id: str, messaging: EnvironmentMessaging) -> Node:
        return _create_router(self._platform, id, messaging)

    def create_port(self, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0,
                    id: str = "") -> Port:
        return _create_port(self._platform, ip, mask, index, id)

    def create_interface(self, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0,
                         id: str = "") -> Interface:
        return _create_interface(self._platform, ip, mask, index, id)

    def create_route(self, net: IPNetwork, port: int, metric: int, id: str = "") -> Route:
        return _create_route(self._platform, net, port, metric, id)

    def add_interface(self, node: Node, interface: Interface, index: int = -1) -> int:
        if node.type == "Router":
            return Router.cast_from(node).add_port(interface.ip, interface.mask, index)
        else:
            return NodeImpl.cast_from(node).add_interface(InterfaceImpl.cast_from(interface))

    def set_interface(self, interface: Interface, ip: Union[str, IPAddress] = "", mask: str = "") -> None:
        iface = InterfaceImpl.cast_from(interface)

        if ip:
            iface.set_ip(ip)

        if mask:
            iface.set_mask(mask)

    def add_service(self, node: Node, *service: Service) -> None:
        node = NodeImpl.cast_from(node)

        for srv in service:
            node.add_service(ServiceImpl.cast_from(srv))

    def remove_service(self, node: Node, *service: Service) -> None:
        node = NodeImpl.cast_from(node)

        for srv in service:
            node.remove_service(ServiceImpl.cast_from(srv))

    def set_shell(self, node: Node, service: Service) -> None:
        NodeImpl.cast_from(node).set_shell(service)

    def add_traffic_processor(self, node: Node, processor: ActiveService) -> None:
        node = NodeImpl.cast_from(node)

        node.add_traffic_processor(processor)

    def add_route(self, node: Node, *route: Route) -> None:
        if node.type != "Router":
            raise RuntimeError("Attempting to add route to non-router node")

        for r in route:
            Router.cast_from(node).add_route(r)

    def add_routing_rule(self, node: Node, rule: FirewallRule) -> None:
        if node.type != "Router":
            raise RuntimeError("Attempting to add route to non-router node")

        Router.cast_from(node).add_routing_rule(rule)

    def set_routing_policy(self, node: Node, policy: FirewallPolicy) -> None:
        if node.type != "Router":
            raise RuntimeError("Attempting to set routing policy to non-router node")

        Router.cast_from(node).set_default_routing_policy(policy)

    def list_routes(self, node: Node) -> List[Route]:
        if node.type != "Router":
            raise RuntimeError("Attempting to add route to non-router node")

        return Router.cast_from(node).list_routes()


# ------------------------------------------------------------------------------------------------------------------
# NodeConfiguration
def _create_node(self: CYSTPlatform, id: str, ip: Union[str, IPAddress] = "", mask: str = "", shell: Service = None) -> Node:
    n = NodeImpl(id, "Node", ip, mask, shell)
    self._general_configuration.add_object(id, n)
    return n


def _create_router(self: CYSTPlatform, id: str, messaging: EnvironmentMessaging) -> Node:
    r = Router(id, messaging)
    self._general_configuration.add_object(id, r)
    return r


def _create_port(self: CYSTPlatform, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0, id: str = "") -> Port:
    if not id:
        id = str(uuid.uuid4())
    i = PortImpl(ip, mask, index, id)
    self._general_configuration.add_object(id, i)
    return i


def _create_interface(self: CYSTPlatform, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0, id: str = "") -> Interface:
    if not id:
        id = str(uuid.uuid4())
    i = InterfaceImpl(ip, mask, index, id)
    self._general_configuration.add_object(id, i)
    return i


def _create_route(self: CYSTPlatform, net: IPNetwork, port: int, metric: int, id: str = "") -> Route:
    if not id:
        id = str(uuid.uuid4())
    r = Route(net, port, metric, id)
    self._general_configuration.add_object(id, r)
    return r



from abc import ABC, abstractmethod
from netaddr import IPAddress, IPNetwork
from typing import NamedTuple, Optional, Tuple, Union

from cyst.api.environment.message import Message
from cyst.api.network.elements import Port, Interface, Connection
from cyst.api.utils.duration import msecs


class Resolver(ABC):
    @abstractmethod
    def resolve_ip(self, id: str, port: int) -> IPAddress:
        pass


class Endpoint:
    def __init__(self, id: str, port: int, ip: Optional[IPAddress] = None):
        self._id = id
        self._port = port
        self._ip = ip

    @property
    def id(self) -> str:
        return self._id

    @property
    def port(self) -> int:
        return self._port

    @property
    def ip(self) -> IPAddress:
        return self._ip

    @ip.setter
    def ip(self, value: IPAddress) -> None:
        self._ip = value

    def __str__(self) -> str:
        return "Endpoint(ID: {}, Port: {}, IP: {})".format(self._id, self._port, self._ip)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: 'Endpoint') -> bool: #type: ignore
        return self.id == other.id and self.port == other.port and self.ip == other.ip


class Hop(NamedTuple):
    src: Endpoint
    dst: Endpoint

    # Necessary for reverse session to make sense
    def swap(self) -> 'Hop':
        return Hop(self.dst, self.src)


class ConnectionImpl(Connection):
    def __init__(self, hop: Optional[Hop] = None) -> None:
        self._hop = hop
        self._blocked = False
        self._delay = 0.0
        self._processing_time = msecs(10).to_float()  # Setting a processing time to a non-zero value

    @property
    def hop(self) -> Hop:
        return self._hop #MYPY: might return optional

    @hop.setter
    def hop(self, value: Hop) -> None:
        self._hop = value

    @property
    def blocked(self) -> bool:
        return self._blocked

    @property
    def delay(self) -> int:
        return self._delay

    def set_params(self, blocked: Optional[bool] = None, delay: Optional[float] = None) -> None:
        if blocked is not None:
            self._blocked = blocked
        if delay is not None:
            self._delay = delay

    def evaluate(self, message: Message) -> Tuple[float, Message]:
        if self.blocked:
            # TODO: return error message
            return -1, message

        return self._processing_time + self._delay, message

    def __str__(self) -> str:
        return f"Connection({self._hop}, Blocked: {self._blocked}, Delay: {self._delay})"

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def cast_from(o: Connection) -> 'ConnectionImpl':
        if isinstance(o, ConnectionImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Connection interface")


class PortImpl(Port):
    def __init__(self, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0, id: str = "") -> None:
        self._id: str = id
        self._ip: Optional[IPAddress] = None
        self._net: Optional[IPNetwork] = None
        self._index: int = index
        self._endpoint: Optional[Endpoint] = None
        self._connection: Optional[Connection] = None

        # Had to use more inelegant check, because IP 0.0.0.0 translates to false
        if ip is not None and ip != "":
            if type(ip) is str:
                self._ip = IPAddress(ip)
            else:
                self._ip = ip

        if mask:
            # Had to use more inelegant check, because IP 0.0.0.0 translates to false
            if ip is None or ip == "":
                raise Exception("Netmask cannot be specified without an IP address")
            if type(ip) is str:
                self._net = IPNetwork(ip + "/" + mask)
            else:
                self._net = IPNetwork(str(ip) + "/" + mask)

    @property
    def ip(self) -> Optional[IPAddress]:
        return self._ip

    def set_ip(self, value: Union[str, IPAddress]) -> None:
        if type(value) is str:
            self._ip = IPAddress(value)
        else:
            self._ip = value

        if self._net:
            # This str dance is sadly necessary, because IPNetwork does not enable changing of IP address
            if type(value) is str:
                self._net = IPNetwork(value + "/" + str(self._net.netmask))
            else:
                self._net = IPNetwork(str(value) + "/" + str(self._net.netmask))

    # Only IP address is returned as an object. Mask is for informative purposes outside construction, so it is
    # returned as a string
    @property
    def mask(self) -> Optional[str]:
        if self._net:
            return str(self._net.netmask)
        else:
            return None

    def set_mask(self, value: str) -> None:
        if not self._ip:
            raise Exception("Netmask cannot be specified without an IP address")

        # This str dance is necessary, because netaddr does not acknowledge changing IPNetwork IP address
        self._net = IPNetwork(str(self._ip) + "/" + value)

    @property
    def net(self) -> Optional[IPNetwork]:
        return self._net

    def set_net(self, value: IPNetwork) -> None:
        self._net = value

    @property
    def endpoint(self) -> Endpoint:
        return self._endpoint #MYPY: might return optional

    # There are no restrictions on connecting an endpoint to the port
    def connect_endpoint(self, endpoint: Endpoint, connection: Connection) -> None:
        self._endpoint = endpoint
        self._connection = connection

    @property
    def index(self) -> int:
        return self._index

    def set_index(self, value: int = 0) -> None:
        self._index = value

    @property
    def connection(self) -> Connection:
        return self._connection

    def set_connection(self, value: Connection) -> None:
        self._connection = value

    # Returns true if given ip belongs to the network
    def routes(self, ip: Union[str, IPAddress] = ""):
        ip = ip if isinstance(ip, IPAddress) else IPAddress(ip)
        return self._net and ip in self._net

    def id(self) -> str:
        return self._id

    @staticmethod
    def cast_from(o: Port) -> 'PortImpl':
        if isinstance(o, PortImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Port interface")


# Interface is just a port, which preserves gateway information (that is a port for end devices)
class InterfaceImpl(PortImpl, Interface):

    def __init__(self, ip: Union[str, IPAddress] = "", mask: str = "", index: int = 0, id: str = ""):
        super(InterfaceImpl, self).__init__(ip, mask, index, id)

        self._gateway_ip: Optional[IPAddress] = None

        if self._ip and self._net:
            # Gateway is by default first host in the network
            self._gateway_ip = next(self._net.iter_hosts())

    def set_ip(self, value: Union[str, IPAddress]) -> None:
        super(InterfaceImpl, self).set_ip(value)

        if self._ip and self._net:
            # Gateway is by default first host in the network
            self._gateway_ip = next(self._net.iter_hosts())

    def set_net(self, value: IPNetwork) -> None:
        super(InterfaceImpl, self).set_net(value)
        self._gateway_ip = next(self._net.iter_hosts())

    def set_mask(self, value: str) -> None:
        super(InterfaceImpl, self).set_mask(value)
        self._gateway_ip = next(self._net.iter_hosts())

    @property
    def gateway(self) -> Optional[IPAddress]:
        return self._gateway_ip

    @property
    def gateway_id(self) -> Optional[str]:
        return self._endpoint.id #MYPY: endpoint might be None

    def connect_gateway(self, ip: IPAddress, connection: Connection, id: str, port: int = 0) -> None:
        if not self._gateway_ip:
            raise Exception("Trying to connect a gateway to an interface without first specifying network parameters")

        if self._gateway_ip != ip:
            raise Exception("Connecting a gateway with wrong configuration")

        self.connect_endpoint(Endpoint(id, port, ip), connection)

    @staticmethod
    def cast_from(o: Interface) -> 'InterfaceImpl':
        if isinstance(o, InterfaceImpl):  #MYPY: Incorrect type in this method and in parent class
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Interface interface")



from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Optional, Union, Dict, Any

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.control import EnvironmentControl
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.platform import PlatformSpecification, Platform
from cyst.api.environment.platform_interface import PlatformInterface


class Environment(ABC):
    """
    The Environment provides a highest-level interface to controlling the simulation. It consists of a number of
    lower-level interfaces that provide a specific functionality.
    """
    @property
    @abstractmethod
    def configuration(self) -> EnvironmentConfiguration:
        """
        This interface is a collection of configuration interfaces for the environment, that are split according to
        their general functionality.

        :rtype: EnvironmentConfiguration
        """

    @property
    @abstractmethod
    def control(self) -> EnvironmentControl:
        """
        This interface provides mechanisms to control the execution of actions within the simulation environment.

        :rtype: EnvironmentControl
        """

    @property
    @abstractmethod
    def messaging(self) -> EnvironmentMessaging:
        """
        This interface enables creating and sending of messages within simulation.

        :rtype: EnvironmentMessaging
        """

    # @property
    # @abstractmethod
    # def policy(self) -> EnvironmentPolicy:
        """
        This environment enables handling of authentication and authorization.

        Warning:
            This interface will be gradually phased off in favor of a new auth framework.

        :rtype: EnvironmentPolicy
        """

    @property
    @abstractmethod
    def resources(self) -> EnvironmentResources:
        """
        This interface gives access to resources, such as actions or exploits.

        :rtype: EnvironmentResources
        """

    @property
    @abstractmethod
    def platform_interface(self) -> PlatformInterface:
        """
        This interface provides means for execution platforms to execute actions and to notify the results of actions.

        :rtype: PlatformInterface
        """

    @property
    @abstractmethod
    def infrastructure(self) -> EnvironmentInfrastructure:
        """
        This interface provides access to environment resources that are aimed at behavioral models and to applications
        utilizing CYST.

        :rtype: EnvironmentInfrastructure
        """

    @property
    @abstractmethod
    def platform(self) -> Platform:
        """
        Returns the underlying execution platform, which is CYST running on.

        :rtype: Platform
        """

    @abstractmethod
    def configure(self, *config_item: ConfigItem, parameters: Dict[str, Any] | None = None) -> 'Environment':
        """
        Configures the environment, according to provided configuration items. This function can be called repeatedly,
        however, each subsequent call replaces the previous configuration. Therefore, a configuration must be done
        at once and every later change in the environment setup must be done through the
        :class:`cyst.api.environment.configuration.EnvironmentConfiguration` interface.

        :param config_item: One or more configuration items. The number of items can be arbitrary and it is not
            order-dependent.
        :type config_item: ConfigItem

        :param parameters: Name-value mapping for configuration that supports parametrization. If none is provided,
            despite parameterizable configuration, default values are used.
        :type parameters: Dict[str, Any] | None

        :return: The configured environment. Used this way for the shorthand form:

        .. code-block:: python

            e = Environment.create().configure(*config, parameters)

        """

    @classmethod
    def create(cls, platform: Optional[Union[str, PlatformSpecification]] = None, run_id: str = "") -> 'Environment':
        """
        Creates a new instance of the environment. A program using CYST can use multiple environments, however, each
        simulation run happens only in the context of one environment.

        :param platform: A specification of a platform to use as a backend for CYST run. By default, a CYST simulation
            using a simulated time is used.
        :type platform: Optional[Union[str, PlatformSpecification]

        :param run_id: The unique id of the current run. If a non-unique id is selected, it may produce unwanted
            results when saving the data to a data store.
        :type run_id: str


        :return: An environment instance.
        """
        import cyst.core.environment.environment
        return cyst.core.environment.environment.create_environment(platform)



from __future__ import annotations

from typing import TYPE_CHECKING, List

from cyst.api.environment.configuration import ExploitConfiguration
from cyst.api.logic.exploit import Exploit, VulnerableService, ExploitLocality, ExploitCategory, ExploitParameter, \
    ExploitParameterType

from cyst.core.environment.stores import ExploitStoreImpl
from cyst.core.logic.exploit import VulnerableServiceImpl, ExploitParameterImpl, ExploitImpl

if TYPE_CHECKING:
    from cyst.core.environment.environment import _Environment


class ExploitConfigurationImpl(ExploitConfiguration):
    def __init__(self, env: _Environment):
        self._env = env

    def create_vulnerable_service(self, id: str, min_version: str = "0.0.0",
                                  max_version: str = "0.0.0") -> VulnerableService:
        return VulnerableServiceImpl(id, min_version, max_version)

    def create_exploit_parameter(self, exploit_type: ExploitParameterType, value: str = "",
                                 immutable: bool = False) -> ExploitParameter:
        return ExploitParameterImpl(exploit_type, value, immutable)

    def create_exploit(self, id: str = "", services: List[VulnerableService] = None, locality: ExploitLocality = ExploitLocality.NONE,
                       category: ExploitCategory = ExploitCategory.NONE,
                       *parameters: ExploitParameter) -> Exploit:
        return ExploitImpl(id, services, locality, category, *parameters)

    def add_exploit(self, *exploits: Exploit) -> None:
        _add_exploit(self._env, *exploits)

    def clear_exploits(self) -> None:
        _clear_exploits(self._env)


# ------------------------------------------------------------------------------------------------------------------
# Exploit configuration
def _add_exploit(self: _Environment, *exploits: Exploit) -> None:
    ExploitStoreImpl.cast_from(self._environment_resources.exploit_store).add_exploit(*exploits)


def _clear_exploits(self: _Environment) -> None:
    ExploitStoreImpl.cast_from(self._environment_resources.exploit_store).clear()



from abc import ABC, abstractmethod
from deprecated.sphinx import versionchanged, versionadded
from netaddr import IPAddress
from typing import Any, Optional, Union, Dict

from cyst.api.environment.message import Message, Request, Response, Status, Signal, ComponentState
from cyst.api.logic.action import Action
from cyst.api.logic.access import Authorization, AuthenticationTarget, AuthenticationToken
from cyst.api.network.session import Session


class EnvironmentMessaging(ABC):
    """
    This interface enables creating and sending of messages within simulation.
    """

    @abstractmethod
    def send_message(self, message: Message, delay: float = 0.0) -> None:
        """
        Sends a message into the simulation for processing. The system will attempt to route the message to its
        destination. If there is a problem with delivery, an indication of it will be received in form of another
        message. This call always succeeds.

        :param message: The message to be sent.
        :type message: Message

        :param delay: Simulation time delay between calling of this function and the actual dispatching of the message
            into the simulated infrastructure.
        :type delay: float

        :return: None
        """

    @abstractmethod
    @versionchanged(version="0.6.0", reason="Added original request parameter to enable request copying")
    def create_request(self, dst_ip: Union[str, IPAddress], dst_service: str = "", action: Optional[Action] = None,
                       session: Optional[Session] = None,
                       auth: Optional[Union[Authorization, AuthenticationToken]] = None,
                       original_request: Optional[Request] = None) -> Request:
        """
        Creates a message of type REQUEST. This function is a domain of active services.

        :param dst_ip: A destination IP address. If provided in the string form, it will attempt to convert it
            internally into the IPAddress.
        :type dst_ip: Union[str, IPAddress]

        :param dst_service: The name of the service this message should be routed to.
        :type dst_service: str

        :param action: An action that should be performed at the destination. Even though this function technically
            enables action-less message, it is an implementation detail and such message would be dropped at the first
            hop. TODO: Check it.
        :type action: Action

        :param session: An optional session to use for a routing of this message.
        :type session: Session

        :param auth: Either authentication or authorization token to use for this message. If an authentication token
            is provided and the type of action is _not_ meta:authenticate, then the system will internally attempt to
            obtain an authorization by means of the meta:authenticate action and the proceed with the original action.
            More in the description of the authentication/authorization framework.
        :type auth: Union[Authorization, AuthenticationToken]

        :param original_request: If provided, values from the original request are used to fill blanks in the new
            request. This is seldom needed in agent's code and is more useful when creating behavioral models.
        :type original_request: Request

        :return: A Request to be sent.
        """

    @abstractmethod
    @versionchanged(version="0.6.0", reason="Added original response parameter to enable response copying")
    def create_response(self, request: Request, status: Status, content: Optional[Any] = None, session: Optional[Session] = None,
                        auth: Optional[Union[Authorization, AuthenticationTarget]] = None,
                        original_response: Optional[Response] = None) -> Response:
        """
        Creates a message of type RESPONSE. This response is always created from a request.

        :param request: The request to which this response is created.
        :type request: Request

        :param status: A status code of the response.
        :type status: Status

        :param content: A message body. Currently anything can be sent in the response, but it is expected that in
            future updates, the content will have its taxonomy and domains, so that automated reasoning can be fully
            applied to message processing.
        :type content: Optional[Any]

        :param session: Either the original session that is inherited from the request, or a new session that is created
            as a part of request processing. If the request arrived through a session, a session must be present with
            the response. Otherwise, the message would not be correctly routed.
        :type session: Session

        :param auth: By convention, this parameter should be the original request's authorization token, unless the
            request resulted in a creation of a new authorization token, or unless the request-response is a part of
            a multi-factor authentication (in which case an Authenticationtarget is returned).
        :type auth: Optional[Union[Authorization, AuthenticationTarget]]

        :param original_response: If provided, values from the original response are used to fill blanks in the new
            request. This is seldom needed in agent's code and is more useful when creating behavioral models.
        :type original_response: Response

        :return: A response to be sent.
        """

    @abstractmethod
    @versionadded(version="0.6.0")
    def create_signal(self, signal_origin: str, state: ComponentState, effect_origin: str,
                      effect_message: Optional[int] = None, effect_description: str = "",
                      effect_parameters: Optional[Dict[str, Any]] = None) -> Signal:
        """
        Creates a message of type SIGNAL.

        :param signal_origin: An identification of a source of the signal. Usually an ID of your component.
        :type signal_origin: str

        :param state: A new state that the component entered.
        :type state: ComponentState

        :param effect_origin: An identification of a source of the state change that prompted the signal emission.
        :type effect_origin: str

        :param effect_message: An ID of a message that caused the state change.
        :type effect_message: Optional[int]

        :param effect_description: A description of an effect used mainly for later analysis.
        :type effect_description: str

        :param effect_parameters: Additional information related to the state change.
        :type effect_parameters: Optional[Dict[str, Any]]
        """

    @abstractmethod
    def open_session(self, request: Request, reverse_direction: bool = False) -> Session:
        """
        Opens a session to the source of the request. If the request already contains a session, it will append the
        session, unless the current node is an endpoint of that session. In that case, the same session is returned.

        :param request: The request from which the session is established.
        :type request: Request
        :param reverse_direction: Whether the direction of the shell is reversed or not.
        :type reverse_direction: bool

        :return: A session to the request source.
        """



from cyst.api.environment.environment import Environment
from cyst.api.configuration import NodeConfig, PassiveServiceConfig, AccessLevel, ExploitConfig, VulnerableServiceConfig
from cyst.api.logic.exploit import ExploitLocality, ExploitCategory

target = NodeConfig(
    active_services=[],
    passive_services=[
        PassiveServiceConfig(
            type="bash",
            owner="root",
            version="8.1.0",
            access_level=AccessLevel.LIMITED,
            local=True,
            id="bash_service"
        ),
        # --------------------------------------------------------------------------------------------------------------
        # Addition to 02_create_machine.py
        # --------------------------------------------------------------------------------------------------------------
        PassiveServiceConfig(
            type="lighttpd",
            owner="www",
            version="1.4.62",
            access_level=AccessLevel.LIMITED,
            local=False,
            id="web_server"
        )
        # --------------------------------------------------------------------------------------------------------------
    ],
    shell="bash",
    interfaces=[],
    traffic_processors=[],
    id="target"
)

# ----------------------------------------------------------------------------------------------------------------------
# Addition to 02_create_machine.py
# ----------------------------------------------------------------------------------------------------------------------
exploit1 = ExploitConfig(
    services=[
        VulnerableServiceConfig(
            name="lighttpd",
            min_version="1.4.62",
            max_version="1.4.62"
        )
    ],
    locality=ExploitLocality.REMOTE,
    category=ExploitCategory.CODE_EXECUTION,
    id="http_exploit"
)

e = Environment.create().configure(target, exploit1)
# ----------------------------------------------------------------------------------------------------------------------
e.control.init()
e.control.run()
e.control.commit()

stats = e.infrastructure.statistics
print(f"Run id: {stats.run_id}\nStart time real: {stats.start_time_real}\n"
      f"End time real: {stats.end_time_real}\nDuration virtual: {stats.end_time_virtual}")



import builtins

from copy import deepcopy
from typing import List, Optional, Dict, Union, Tuple, Any

from cyst.api.environment.configuration import RuntimeConfiguration
from cyst.api.environment.environment import EnvironmentMessaging
from cyst.api.environment.message import Message
from cyst.api.environment.stores import ActionStore, ExploitStore, ServiceStore
from cyst.api.environment.platform_specification import PlatformSpecification, PlatformType
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.logic.access import AccessLevel
from cyst.api.logic.action import ActionDescription, Action
from cyst.api.logic.exploit import Exploit, ExploitCategory, ExploitLocality
from cyst.api.host.service import PassiveService, ActiveServiceDescription, ActiveService
from cyst.api.network.node import Node

from cyst.core.logic.action import ActionImpl
# from cyst.core.host.service import ServiceImpl
# from cyst.core.network.session import SessionImpl
# from cyst.core.network.node import NodeImpl


class ServiceStoreImpl(ServiceStore):

    def __init__(self, messaging: EnvironmentMessaging, resources: EnvironmentResources, runtime_configuration: RuntimeConfiguration):
        self._services = {}
        self._messaging = messaging
        self._resources = resources
        self._runtime_configuration = runtime_configuration

        self._service_instances = {}
        self._service_instance_ids = {}

    def add_service(self, description: ActiveServiceDescription) -> None:
        self._services[description.name] = description

    def get_service(self, name: str) -> Optional[ActiveServiceDescription]:
        if name in self._services:
            return self._services[name]
        else:
            return None

    def create_active_service(self, type: str, owner: str, name: str, node: Node,
                              service_access_level: AccessLevel = AccessLevel.LIMITED,
                              configuration: Optional[Dict[str, Any]] = None, id: str = "") -> Optional[ActiveService]:
        if not type in self._services:
            raise RuntimeError(f"Attempting to create nonexistent service {type}.")
        service_description: ActiveServiceDescription = self._services[type]

        # Configuration is extended by parameters that are passed through the runtime configuration mechanism.
        # The runtime configuration takes precedence over what is inside the configuration file.
        if not configuration:
            configuration = {}

        configuration.update(self._runtime_configuration.other_params)

        service = service_description.creation_fn(self._messaging, self._resources, id, configuration)
        #return ServiceImpl(type, service, name, owner, service_access_level, id)

        # service is stored according to its in-engine ID as well as its address to help with message sending
        self._service_instances[id] = service
        self._service_instance_ids[builtins.id(service)] = id

        return service

    def get_active_service(self, id) -> Optional[ActiveService]:
        return self._service_instances[id]


    def get_active_services(self) -> List[ActiveService]:
        return list(self._service_instances.values())

    def get_active_service_id(self, address: str) -> str:
        return self._service_instance_ids[address]

class ActionStoreImpl(ActionStore):

    def __init__(self, platform: Optional[PlatformSpecification] = None):
        self._actions = {}
        if platform:
            self._platform = platform
        else:
            self._platform = PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST")

    def get(self, id: str = "") -> Optional[Action]:
        if id in self._actions:
            return deepcopy(self._actions[id])
        return None

    def get_ref(self, id: str = "") -> Optional[Action]:
        if id in self._actions:
            return self._actions[id]
        return None

    def get_prefixed(self, prefix: str = "") -> List[Action]:
        result = []
        for id, value in self._actions.items():
            if id.startswith(prefix):
                result.append(deepcopy(value))
        return result

    def add(self, action: ActionDescription) -> None:
        action_platform = action.platform

        if not action_platform:
            action_platform = PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST")

        if type(action_platform) != list:
            action_platform = [action_platform]

        if not self._platform in action_platform:
            return

        if action.id in self._actions:
            raise RuntimeError(f"Attempting to add action with ID that is already present for platform {self._platform}: {action.id}")
        self._actions[action.id] = ActionImpl(action)


class ExploitStoreImpl(ExploitStore):
    def __init__(self):
        # Nested lists here are to allow easier conversion to sets and to do the intersection
        # TODO: revise the data structure logic here
        self._by_id: Dict[str, List[Exploit]] = {}
        self._by_service: Dict[str, List[Exploit]] = {}
        self._by_category: Dict[ExploitCategory, List[Exploit]] = {}

    def clear(self) -> None:
        self._by_id.clear()
        self._by_service.clear()
        self._by_category.clear()

    def add_exploit(self, *exploits: Exploit) -> None:
        for exploit in exploits:
            # Exploit already in store, do nothing
            if exploit.id in self._by_id:
                return

            self._by_id[exploit.id] = [exploit]

            for service in exploit.services.values():
                if service.id not in self._by_service:
                    self._by_service[service.id] = []

                self._by_service[service.id].append(exploit)

            if exploit.category not in self._by_category:
                self._by_category[exploit.category] = []

            self._by_category[exploit.category].append(exploit)

    def get_exploit(self, id: str = "", service: str = "", category: ExploitCategory = ExploitCategory.NONE) -> Optional[List[Exploit]]:
        candidate_sets = []

        if id:
            candidate_sets.append(set(self._by_id.get(id, [])))

        if service:
            candidate_sets.append(set(self._by_service.get(service, [])))

        if category != ExploitCategory.NONE:
            candidate_sets.append(set(self._by_category.get(category, [])))

        if not candidate_sets:
            result = []
            for exploit_list in self._by_id.values():
                result.extend(exploit_list)
            return result
        else:
            return list(set.intersection(*candidate_sets))

    def evaluate_exploit(self, exploit: Union[str, Exploit], message: Message, node: Node) -> Tuple[bool, str]:
        if isinstance(exploit, str):
            exploits = self.get_exploit(exploit)
            if not exploits:
                return False, "Could not find exploit by id"
            else:
                # Gah!
                exploit = exploits[0]

        # For exploit to be applicable, a number of conditions must be satisfied.
        # 1) Local exploits can only be used at the session end
        if exploit.locality == ExploitLocality.LOCAL:
            if not message.session:
                return False, "Local exploits can only be used from within an existing session"
            elif not message.session.end[0] in [iface.ip for iface in node.interfaces]:
                return False, "Local exploits can only be used at session endpoint."

        # 2) Exploits must be applicable on given service_id
        exploit_service = exploit.services.get(message.dst_service, None)
        if not exploit_service:
            return False, "Attempting to use exploit on a service it does not apply to"

        # 3) service_id must exist on node
        node_service = node.services.get(message.dst_service, None)
        if not node_service:
            return False, "The exploit is not fit for any of the services on the node"

        # 4) Versions must match
        if isinstance(node_service, PassiveService) and not exploit_service.min_version <= node_service.version <= exploit_service.max_version:
            return False, "The exploit is not applicable on the current version of a service"

        # 5) TODO There will also be tags to evaluate, but that is a future endeavour

        return True, ""

    @staticmethod
    def cast_from(o: ExploitStore) -> 'ExploitStoreImpl':
        if isinstance(o, ExploitStoreImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the ExploitStore interface")



from dataclasses import dataclass, field
from typing import List, Union, Optional
from uuid import uuid4
from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration.host.service import ActiveServiceConfig
from cyst.api.configuration.network.elements import InterfaceConfig, RouteConfig
from cyst.api.configuration.network.firewall import FirewallConfig


@serialize(type_check=coerce)
@dataclass
class RouterConfig(ConfigItem):
    """ Configuration of a network router

    Router models an active network device that forwards messages over the network according to rules. At the conceptual
    level it conflates the concept of switch and router to one device.

    Currently, a router is implemented as a special type of node, with distinct code paths, in the future it is expected
    that a router would be implemented as an active service on a node. This will enable better logical separation
    between its routing, firewalling, and IDS/IPS activities. It will also be more fit for SDN modelling.

    :param interfaces: A list of network interfaces.
    :type interfaces: List[Union[InterfaceConfig, str]]

    :param traffic_processors: A list of either active services that are acting as traffic processors, or at most one
        firewall. Firewall is used as a mechanism for router to do inter-network routing, by means of a FORWARD chain.
    :type traffic_processors: Optional[List[Union[FirewallConfig, ActiveServiceConfig, str]]]

    :param routing_table: A routing configuration for inter-router communication. Routing to end devices is arranged
        automatically when creating connections between end devices and the router. Networks are inferred from
        interface configurations.
    :type routing_table: List[RouteConfig]

    :param firewall: A configuration of firewall rules.
    :type firewall: Optional[FirewallConfig]
    """
    interfaces: List[Union[InterfaceConfig, str]]
    traffic_processors: List[Union[FirewallConfig, ActiveServiceConfig, str]]
    routing_table: List[RouteConfig] = field(default_factory=list)  # TODO: check if such a default is ok
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__router"
    id: str = ""




from typing import Any, Dict, List, NamedTuple, Optional, Tuple
from netaddr import IPAddress, IPNetwork

from cyst.api.host.service import ActiveService, ActiveServiceDescription
from cyst.api.environment.message import Message, Request, Status, StatusValue, StatusOrigin, Signal
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.network.firewall import Firewall, FirewallChainType, FirewallRule, FirewallPolicy


class FirewallImpl(ActiveService, Firewall):

    class FirewallChain:
        def __init__(self, type: FirewallChainType, policy: FirewallPolicy = FirewallPolicy.DENY) -> None:
            self._type = type
            self._default_policy = policy
            self._rules: List[FirewallRule] = []

        @property
        def default_policy(self) -> FirewallPolicy:
            return self._default_policy

        @property
        def type(self) -> FirewallChainType:
            return self._type

        def set_default_policy(self, policy: FirewallPolicy) -> None:
            self._default_policy = policy

        def add_rule(self, rule: FirewallRule) -> None:
            self._rules.append(rule)

        def remove_rule(self, index: int) -> None:
            del self._rules[index]

        def list_rules(self) -> List[FirewallRule]:
            return self._rules

        def evaluate(self, src_ip: IPAddress, dst_ip: IPAddress, dst_service: str) -> Tuple[bool, int]:
            for rule in self._rules:
                if src_ip in rule.src_net and dst_ip in rule.dst_net and (rule.service == "*" or dst_service == rule.service):
                    if rule.policy == FirewallPolicy.ALLOW:
                        return True, 0
                    elif rule.policy == FirewallPolicy.DENY:
                        return False, 0

            if self._default_policy == FirewallPolicy.ALLOW:
                return True, 0

            return False, 0

    def __init__(self, env: EnvironmentMessaging = None, args: Optional[Dict[str, Any]] = None) -> None:
        self._msg = env
        default_policy = FirewallPolicy.DENY
        if args:
            default_policy = args.get("default_policy", FirewallPolicy.DENY)

        self._chains = {
            FirewallChainType.INPUT: FirewallImpl.FirewallChain(FirewallChainType.INPUT, default_policy),
            FirewallChainType.OUTPUT: FirewallImpl.FirewallChain(FirewallChainType.OUTPUT, default_policy),
            FirewallChainType.FORWARD: FirewallImpl.FirewallChain(FirewallChainType.FORWARD, default_policy)
        }

        self._local_ips: List[IPAddress] = []

    async def run(self):
        pass

    def add_local_ip(self, ip: IPAddress) -> None:
        if not ip in self._local_ips:
            self._local_ips.append(ip)

    def remove_local_ip(self, ip: IPAddress) -> None:
        self._local_ips.remove(ip)

    def add_rule(self, chain: FirewallChainType, rule: FirewallRule) -> None:
        self._chains[chain].add_rule(rule)

    def remove_rule(self, chain: FirewallChainType, index: int) -> None:
        self._chains[chain].remove_rule(index)

    def list_rules(self, chain: Optional[FirewallChainType] = None) -> List[Tuple[FirewallChainType, FirewallPolicy, List[FirewallRule]]]:
        if not chain:
            return [(x.type, x.default_policy, x.list_rules()) for x in self._chains.values()]

        return NotImplemented  #MYPY:  Case not Handled

    def set_default_policy(self, chain: FirewallChainType, policy: FirewallPolicy) -> None:
        self._chains[chain].set_default_policy(policy)

    def get_default_policy(self, chain: FirewallChainType) -> FirewallPolicy:
        return self._chains[chain].default_policy

    def evaluate(self, src_ip: IPAddress, dst_ip: IPAddress, dst_service: str) -> Tuple[bool, int]:
        local_src = src_ip in self._local_ips
        local_dst = dst_ip in self._local_ips

        # By convention allow all communication between local IPs (usually loopback)
        chain = None
        if local_src and local_dst:
            return True, 0
        elif local_src:
            chain = FirewallChainType.OUTPUT
        elif local_dst:
            chain = FirewallChainType.INPUT
        else:
            chain = FirewallChainType.FORWARD

        return self._chains[chain].evaluate(src_ip, dst_ip, dst_service)

    async def process_message(self, message: Message) -> Tuple[bool, int]:
        # Firewalls are active services, so they receive signals as well. We just ignore them.
        if isinstance(message, Signal):
            return True, 0

        result, processing_time = self.evaluate(message.src_ip, message.dst_ip, message.dst_service)

        # Sending a response that the message did not arrive
        if not result:
            assert(isinstance(message, Request))
            response = self._msg.create_response(message, Status(StatusOrigin.NETWORK, StatusValue.FAILURE),
                                                 f"Destination {message.dst_ip}:{message.dst_service} unreachable",
                                                 session=message.session, auth=message.auth)
            self._msg.send_message(response, processing_time)

        return result, processing_time


def create_firewall(msg: EnvironmentMessaging, res: EnvironmentResources, id: str, args: Optional[Dict[str, Any]]) -> ActiveService:
    fw = FirewallImpl(msg, args)
    return fw


service_description = ActiveServiceDescription(
    "firewall",
    "An implementation of a firewall service",
    create_firewall
)


import asyncio
import logging

from typing import Dict, Tuple
from uuid import uuid4

from cyst.api.environment.data_model import ActionModel
from cyst.api.environment.stores import DataStore
from cyst.api.environment.message import Request, Response, Message, Timeout
from cyst.api.environment.configuration import GeneralConfiguration
from cyst.api.host.service import ActiveService, Service
from cyst.api.logic.behavioral_model import BehavioralModel
from cyst.api.logic.composite_action import CompositeActionManager

from cyst.core.environment.environment import EnvironmentMessagingImpl, EnvironmentResources
from cyst.core.logic.action import ActionImpl, ActionType


class CompositeActionManagerImpl(CompositeActionManager):
    def __init__(self, loop: asyncio.AbstractEventLoop, behavioral_models: Dict[str, BehavioralModel],
                 messaging: EnvironmentMessagingImpl, resources: EnvironmentResources, general: GeneralConfiguration,
                 data_store: DataStore, active_actions: Dict[str, ActionModel]):
        self._loop = loop
        self._futures = {}
        self._incoming_queue = set()
        self._outgoing_queue = set()
        self._composite_queue = set()
        self._coroutines = {}
        self._models = behavioral_models
        self._msg = messaging
        self._res = resources
        self._general = general
        self._processing = False
        self._messages = set()
        self._set_futures = 0
        self._composites_processing = 0
        self._terminate = False
        self._log = logging.getLogger("system")
        self._subordinate_action_counts = {}
        self._task_parents = {}
        self._composites_processing = 0
        self._data_store = data_store
        self._active_actions = active_actions

        self._loop.set_task_factory(self.task_factory)

    def clear_resource_task(self, task: asyncio.Task):
        task_name = task.get_name()
        parent_task = self._task_parents[task_name]
        self._subordinate_action_counts[parent_task] -= 1

    def task_factory(self, loop: asyncio.AbstractEventLoop, coro, context = None) -> asyncio.Task:
        task = asyncio.Task(coro=coro, loop=loop)

        # We really hijack the task creation only to get access to call_action parent
        if coro.cr_code.co_name == "call_action":
            task_name = asyncio.current_task().get_name()
            self._task_parents[task.get_name()] = task_name
            self._subordinate_action_counts[task_name] += 1

        # External resource tasks are a bit different, mainly because we need to take care of them only if they are
        # executed in the context of a composite action
        if coro.cr_code.co_name == "send" or coro.cr_code.co_name == "receive":
            task_name = asyncio.current_task().get_name()
            if task_name in self._subordinate_action_counts:
                self._task_parents[task.get_name()] = task_name
                self._subordinate_action_counts[task_name] += 1
                task.add_done_callback(self.clear_resource_task)

        return task

    def execute_request(self, request: Request, delay: int) -> None:
        self._composite_queue.add(request)
        self._coroutines[request.id] = self._models[request.action.namespace].action_flow(request)

    async def call_action(self, request: Request, delay: float = 0.0) -> None:
        task_name = asyncio.current_task().get_name()

        if task_name in self._subordinate_action_counts:
            self._subordinate_action_counts[task_name] += 1

        future = self._loop.create_future()
        self._futures[request.id] = future

        if ActionImpl.cast_from(request.action).type == ActionType.COMPOSITE:
            self._composite_queue.add(request)
            self._coroutines[request.id] = self._models[request.action.namespace].action_flow(request)
        else:
            self._outgoing_queue.add(request)

        await future

        if task_name in self._task_parents:
            self._subordinate_action_counts[self._task_parents[task_name]] -= 1
            del self._task_parents[task_name]
        else:
            self._subordinate_action_counts[task_name] -= 1

        return future.result()

    async def delay(self, delay: float = 0.0) -> None:
        future = self._loop.create_future()
        future_id = uuid4()
        self._futures[future_id] = future

        self._res.clock.timeout(self._process_timeout, delay, future_id)

        task_name = asyncio.current_task().get_name()
        if task_name in self._subordinate_action_counts:
            self._subordinate_action_counts[task_name] += 1

        await future
        return future.result()

    def _process_timeout(self, message: Message) -> Tuple[bool, int]:
        if isinstance(message, Timeout):
            self._futures[message.parameter].set_result(message.start_time + message.duration)
        # Timeout is processed instantly
        return True, 0

    def is_composite(self, id: int) -> bool:
        return id in self._messages

    def incoming_message(self, message: Message) -> None:
        self._incoming_queue.add(message)

    async def send_request(self, request: Request):
        self._log.debug(f"[start] Composite action: sending message with id {request.id}")
        self._messages.add(request.id)
        self._msg.send_message(request)
        self._log.debug(f"[ end ] Composite action: sending message with id {request.id}")

    async def process_composite(self, request) -> None:
        self._log.debug(f"[start] Composite action: processing request from composite queue: {request}")

        task_name = asyncio.current_task().get_name()
        self._subordinate_action_counts[task_name] = 0

        delay, response = await self._coroutines[request.id]
        del self._coroutines[request.id]

        # If the complex message was the top-level one, return the result to a service
        if request.id in self._futures:
            self._futures[request.id].set_result(response)
        else:
            caller_id = request.platform_specific["caller_id"]
            service = self._general.get_object_by_id(caller_id, ActiveService)
            await service.process_message(response)
            if response.id in self._active_actions:
                active_action = self._active_actions[response.id]
                active_action.set_response(response)
                self._data_store.add_action(active_action)
                del self._active_actions[response.id]

        self._composites_processing -= 1
        self._log.debug(f"[ end ] Composite action: processing request from composite queue. Got this response: {response}")
        del self._subordinate_action_counts[asyncio.current_task().get_name()]

    async def process(self) -> Tuple[bool, bool, bool]:
        while self._composite_queue:
            request = self._composite_queue.pop()
            self._loop.create_task(self.process_composite(request))
            self._composites_processing += 1

        while self._outgoing_queue:
            request = self._outgoing_queue.pop()
            self._log.debug(f"[start] Composite action: processing request from outgoing queue: {request}")
            await self._loop.create_task(self.send_request(request))
            self._log.debug(f"[ end ] Composite action: processing request from outgoing queue: {request}")

        while self._incoming_queue:
            response = self._incoming_queue.pop()
            self._log.debug(f"[start] Composite queue: processing response from incoming queue: {response}")
            self._futures[response.id].set_result(response)
            self._log.debug(f"[ end ] Composite queue: processing response from incoming queue: {response}")

        # This is more important than it looks!
        if self._composites_processing > 0:
            # Yield control to composite tasks processing, just to be sure it does not get starved.
            await asyncio.sleep(0)

        # Indicate we have something to process if there is anything in queues. By the nature of the code above, when
        # these queues are empty then everything should either be resolved or stuck in an await.
        composite_actions_resolving = False
        for actions_remaining in self._subordinate_action_counts.values():
            if actions_remaining == 0:
                composite_actions_resolving = True
                break

        return bool(self._composite_queue) or bool(self._outgoing_queue) or bool(self._incoming_queue), composite_actions_resolving, self._composites_processing > 0



from abc import ABC, abstractmethod
from datetime import datetime
from deprecated.sphinx import versionadded, versionchanged
from typing import Any, Union, Callable, Tuple

from cyst.api.environment.message import Message
from cyst.api.host.service import ActiveService


class Clock(ABC):
    """ Clock interface provides access to time management of a given platform.
    """
    @versionadded(version="0.6.0")
    @abstractmethod
    def current_time(self) -> float:
        """ Returns a current time of a platform as an offset from a specific time in the past. In case of a discrete
        simulation platform, it will likely return a whole number. For other environments, this will be a fractional
        number (aka. python's time() function).

        :return: Current time as an offset.
        """

    @versionadded(version="0.6.0")
    @abstractmethod
    def real_time(self) -> datetime:
        """ Returns a current time of a platform converted to a real date-time information. In case of a discrete
        simulation platform this entails a conversion of time offset to a real time. In case o a real-time environment
        this will likely be only a reading of a system clock.

        :return: Current time as a datetime structure.
        """

    @versionchanged(version="0.6.0", reason="Changed delay type to float and added support for generic callbacks.")
    @abstractmethod
    def timeout(self, callback: Union[ActiveService, Callable[[Message], Tuple[bool, int]]], delay: float, parameter: Any = None) -> None:
        """ Schedule a timeout message. This acts like a time callback and enables inclusion of any kind of data.

        :param callback: Either a service, which should receive the timeout message, or an arbitrary callback.
        :param delay: The duration of the timeout in simulation time.
        :param parameter: The included data. They will not be modified.
        :return: None
        """
